"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from nagini_translation.lib.constants import (
    BYTES_TYPE,
    COMBINED_NAME_ACCESSOR,
    DICT_TYPE,
    END_LABEL,
    IGNORED_IMPORTS,
    IGNORED_MODULE_NAMES,
    INT_TYPE,
    LIST_TYPE,
    MAY_SET_PRED,
    NAME_QUANTIFIER_VAR,
    NAME_DOMAIN,
    OBJECT_TYPE,
    PRIMITIVES,
    RANGE_TYPE,
    SET_TYPE,
    TUPLE_TYPE,
)
from nagini_translation.lib.program_nodes import (
    GenericType,
    OptionalType,
    PythonField,
    PythonGlobalVar,
    PythonMethod,
    PythonModule,
    PythonNode,
    PythonTryBlock,
    PythonType,
    PythonVar,
    SilverType,
    UnionType,
    toposort_classes,
    chain_if_stmts,
)
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Position,
    Stmt,
    StmtsAndExpr
)
from nagini_translation.lib.util import (
    AssignCollector,
    contains_stmt,
    flatten,
    get_body_indices,
    get_parent_of_type,
    get_surrounding_try_blocks,
    InvalidProgramException,
    is_get_ghost_output,
    UnsupportedException,
)
from nagini_translation.translators.abstract import Context
from nagini_translation.translators.common import CommonTranslator
from typing import List, Optional, Tuple, Union


class StatementTranslator(CommonTranslator):
    def __init__(self, config: 'TranslatorConfig', jvm: 'JVM', source_file: str,
                 type_info: 'TypeInfo', viper_ast: 'ViperAST') -> None:
        super().__init__(config, jvm, source_file, type_info, viper_ast)
        # Keep track of the end and after labels of loops we are currently in.
        self.loops = {}
        self.imported_modules = set()

    def translate_stmt(self, node: ast.AST, ctx: Context) -> List[Stmt]:
        """
        Generic visitor function for translating statements
        """
        method = 'translate_stmt_' + node.__class__.__name__
        visitor = getattr(self, method, self.translate_generic)
        return visitor(node, ctx)

    def _execute_module_statements(self, module: PythonModule, import_stmt: ast.AST,
                                   ctx: Context) -> Stmt:
        """
        Creates a single statement that represents the execution of all global statements
        in the given module (including those of other imported modules), if said module
        has not been imported before.
        """
        if ctx.module not in self.imported_modules:
            self.imported_modules.add(ctx.module)
        pos = self.to_position(import_stmt, ctx)
        info = self.no_info(ctx)
        cond_stmts = []
        if module in self.imported_modules:
            return self.viper.Assert(module.defined_var[1], pos, info)
        set_defined = self.viper.LocalVarAssign(module.defined_var[1],
                                                self.viper.TrueLit(pos, info),
                                                pos, info)
        cond_stmts.append(set_defined)
        old_module = ctx.module
        old_try_blocks = ctx.current_function.try_blocks
        old_try_labels = ctx.current_function.labels
        old_try_precondition = ctx.current_function.precondition
        old_try_postcondition = ctx.current_function.postcondition
        old_try_loop_invariants = ctx.current_function.loop_invariants

        ctx.current_function.try_blocks = module.try_blocks
        ctx.current_function.labels = module.labels
        ctx.current_function.precondition = module.precondition
        ctx.current_function.postcondition = module.postcondition
        ctx.current_function.loop_invariants = module.loop_invariants
        ctx.current_function._module = module
        ctx.module = module
        self.imported_modules.add(module)

        old_label_aliases = ctx.label_aliases.copy()
        # Create label aliases
        for label in module.labels:
            new_label = ctx.current_function.get_fresh_name(label)
            ctx.label_aliases[label] = new_label

        ctx.added_handlers.append((module, ctx.var_aliases, ctx.label_aliases))

        for stmt in module.node.body:
            cond_stmts.extend(self.translate_stmt(stmt, ctx))

        ctx.label_aliases = old_label_aliases

        self.imported_modules.remove(module)
        ctx.module = old_module
        ctx.current_function._module = old_module
        ctx.current_function.try_blocks = old_try_blocks
        ctx.current_function.labels = old_try_labels
        ctx.current_function.precondition = old_try_precondition
        ctx.current_function.postcondition = old_try_postcondition
        ctx.current_function.loop_invariants = old_try_loop_invariants

        not_defined = self.viper.Not(module.defined_var[1], pos, info)
        cond = self.viper.If(not_defined, self.translate_block(cond_stmts, pos, info),
                             self.translate_block([], pos, info), pos, info)
        return cond

    def translate_stmt_ImportFrom(self, node: ast.ImportFrom, ctx: Context) -> List[Stmt]:
        """
        A global ImportFrom is translated to statements that 1) execute the statements
        in the imported module if the module has not been imported before, 2a) if a list
        of specific names are imported, assert they are available in the specified module
        and then add them to the current module, 2b) if all names are imported, the names
        in the imported module are simply added to the current module.
        """
        if not self.is_main_method(ctx):
            raise InvalidProgramException(node, 'local.import')
        stmts = []
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)

        for imported in ctx.module.from_imports:
            if not isinstance(imported, PythonModule):
                imported = imported.original_module
            if (imported.type_prefix == node.module or
                    imported.file.startswith(node.module)):
                mod = imported
                break
        else:
            assert node.module in IGNORED_IMPORTS

        if node.module not in IGNORED_IMPORTS:
            stmts.append(self._execute_module_statements(mod, node, ctx))

        import_all = False
        imported_names = []

        if len(node.names) == 1 and node.names[0].name == '*':
            if node.module not in IGNORED_IMPORTS:
                import_all = True
            else:
                imported_names.extend([(n, n) for n in IGNORED_MODULE_NAMES[node.module]])

        else:
            imported_names.extend([(n.name, n.asname if n.asname else n.name)
                                   for n in node.names])

        if import_all:
            union = self.viper.AnySetUnion(ctx.module.names_var[1], mod.names_var[1],
                                           pos, info)
            stmts.append(
                self.viper.LocalVarAssign(ctx.module.names_var[1], union, pos, info))
        else:
            for alias in imported_names:
                name = alias[0]
                as_name = alias[1]
                msg = 'name "' + name + '" is defined in imported module'
                pos = self.to_position(node, ctx, error_string=msg)
                if node.module not in IGNORED_IMPORTS:
                    name_int = self.viper.IntLit(self._get_string_value(name), pos, info)
                    exists_in_other = self._is_defined(name_int, imported.names_var[1],
                                                       pos, info)
                    # Make sure the name is actually defined.
                    stmts.append(self.viper.Assert(exists_in_other, pos, info))
                as_name_int = self.viper.IntLit(self._get_string_value(as_name), pos,
                                                info)
                stmts.append(self._set_global_defined(as_name_int,
                                                      ctx.module.names_var[1], pos, info))

        return stmts

    def _get_name_from_combined(self, e: Expr, pos: Position, info: Info) -> Expr:
        """
        Assuming that the given expression represents a name that is a combination of
        some prefix and a local name, returns the local name.
        """
        return self.viper.DomainFuncApp(COMBINED_NAME_ACCESSOR, [e], self.name_type(),
                                        pos, info, NAME_DOMAIN)

    def translate_stmt_Import(self, node: ast.Import, ctx: Context) -> List[Stmt]:
        """
        A global Import is translated to statements that 1) execute the statements
        in the imported module if the module has not been imported before, 2) add all
        names in the imported module, combined with the module name as a prefix, to the
        current module.
        """
        # TODO: What we're doing here does not take into account names that are added to
        # a module after it's imported; they will not be available from the module that
        # contains the import. This should be okay since a) if anything, it would be
        # incomplete, not unsound, and b) for non-cyclic imports, all names should be
        # defined in the imported module after its statements have been executed, and
        # for cyclic imports, more names may be added, but only after all statements that
        # might use them have been executed, so it should no longer be relevant.
        if not self.is_main_method(ctx):
            raise InvalidProgramException(node, 'local.import')
        stmts = []
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        for name in node.names:
            if name.name in IGNORED_IMPORTS:
                continue
            imported_name = name.asname if name.asname else name.name
            name_parts = imported_name.split('.')
            mod = ctx.module
            for part in name_parts:
                mod = mod.namespaces[part]
            module_set_type = SilverType(self.viper.SetType(self.name_type()), ctx.module)
            new_set_var = ctx.current_function.create_variable('new_set',
                                                               module_set_type,
                                                               self.translator)
            stmts.append(self._execute_module_statements(mod, node, ctx))
            name_var_decl = self.viper.LocalVarDecl(NAME_QUANTIFIER_VAR, self.name_type(),
                                                    pos, info)
            name_var_ref = self.viper.LocalVar(NAME_QUANTIFIER_VAR, self.name_type(), pos,
                                               info)
            name_in_imported = self._is_defined(name_var_ref, mod.names_var[1], pos, info)


            combined_name = name_var_ref
            last_part = name_var_ref
            name_ints = []
            for name in reversed(name_parts):
                name_int = self.viper.IntLit(self._get_string_value(name), pos, info)
                combined_name = self._combine_names(name_int, combined_name, pos, info)
                last_part = self._get_name_from_combined(last_part, pos, info)
                name_ints.append(name_int)
            combined_part = last_part
            for name_int in name_ints:
                combined_part = self._combine_names(name_int, combined_part, pos, info)
            combined_in_new = self._is_defined(combined_name, new_set_var.ref(), pos,
                                               info)
            impl = self.viper.EqCmp(name_in_imported, combined_in_new, pos, info)
            trigger = self.viper.Trigger([combined_in_new], pos, info)
            assertion = self.viper.Forall([name_var_decl], [trigger], impl, pos, info)
            stmts.append(self.viper.Inhale(assertion, pos, info))
            union = self.viper.AnySetUnion(ctx.module.names_var[1], new_set_var.ref(),
                                           pos, info)
            stmts.append(self.viper.LocalVarAssign(ctx.module.names_var[1], union, pos,
                                                   info))
        return stmts

    def translate_stmt_FunctionDef(self, node: ast.FunctionDef,
                                   ctx: Context) -> List[Stmt]:
        """
        Function definitions in the global method are translated to a check that all
        dependencies of the declaration are defined, and subsequently an assignment
        that sets the function name to be defined.
        """
        assert self.is_main_method(ctx)
        if ctx.current_class:
            method = ctx.current_class.get_func_or_method(node.name)
        else:
            method = ctx.module.get_func_or_method(node.name)
        if not method:
            method = ctx.module.predicates.get(node.name)
        if not method:
            method = ctx.module.io_operations.get(node.name)
        if not method:
            return []
        dep_check = self._check_dependencies_defined(method, node, ctx)
        return dep_check + [self.set_global_defined(method, ctx.module, node, ctx)]

    def translate_stmt_ClassDef(self, node: ast.ClassDef, ctx: Context) -> List[Stmt]:
        """
        Class definitions in the global method are translated to 1) a check that
        the dependencies are defined, 2) translations of static field assignments and 3)
        an assignment that sets the class name to be defined.
        """
        assert self.is_main_method(ctx)
        # static field definitions
        cls = ctx.module.classes[node.name]
        stmts = []
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)

        stmts.extend(self._check_dependencies_defined(cls, node, ctx))
        for base in node.bases:
            decl = self.get_target(base, ctx)
            if (isinstance(decl, PythonType) and
                        decl.python_class.module is not ctx.module.global_module):
                stmts.extend(self.assert_global_defined(decl, ctx.module, base, ctx,
                                                        call_deps=False))

        ctx.current_class = cls
        full_perm = self.viper.FullPerm(pos, info)
        for field in cls.static_fields.values():
            if not field.is_final:
                field_acc = self.translate_static_field_access(field, cls, field.node,
                                                               ctx)
                acc_pred = self.viper.FieldAccessPredicate(field_acc, full_perm, pos,
                                                           info)
                stmts.append(self.viper.Inhale(acc_pred, pos, info))

        for stmt in node.body:
            stmts.extend(self.translate_stmt(stmt, ctx))
        ctx.current_class = None
        return stmts + [self.set_global_defined(cls, ctx.module, node, ctx)]

    def _check_dependencies_defined(self, py_node: PythonNode, node: ast.AST,
                                    ctx: Context) -> List[Stmt]:
        """
        Returns statements that assert that all dependencies needed for the declaration
        of the given PythonNode are currently defined.
        """
        if ctx.current_class:
            return []
        msg = 'all dependencies of "' + node.name + '" are defined'
        dep_pos = self.to_position(node, ctx, error_string=msg)
        info = self.no_info(ctx)

        deps_defined = self.viper.TrueLit(dep_pos, info)
        for ref, decl, mod in py_node.definition_deps:
            module_set = mod.names_var[1]
            decl_ids = self.extract_identifiers(ref, dep_pos, info)
            for decl_id in decl_ids:
                contains = self._is_defined(decl_id, module_set, dep_pos, info)
                deps_defined = self.viper.And(deps_defined, contains, dep_pos, info)
        return [self.viper.Assert(deps_defined, dep_pos, info)]

    def translate_stmt_Global(self, node: ast.Global, ctx: Context) -> List[Stmt]:
        # No need to do anything, this just signals what variables refer to.
        return []

    def translate_stmt_Delete(self, node: ast.Delete, ctx: Context) -> List[Stmt]:
        result = []
        info = self.no_info(ctx)
        full_perm = self.viper.FullPerm(self.no_position(ctx), info)
        for t in node.targets:
            target = self.get_target(t, ctx)
            if isinstance(target, PythonField):
                pos = self.to_position(t, ctx)
                # assume t is an ast.Attribute
                stmts, receiver = self.translate_expr(t.value, ctx)
                result.extend(stmts)
                rec_type = self.get_type(t.value, ctx)
                python_field = rec_type.get_field(t.attr)
                field = self.viper.Field(python_field.sil_name,
                                         self.translate_type(python_field.type, ctx),
                                         pos, info)
                field_acc = self.viper.FieldAccess(receiver, field, pos, info)
                field_acc_pred = self.viper.FieldAccessPredicate(field_acc, full_perm,
                                                                 pos, info)
                result.append(self.viper.Exhale(field_acc_pred, pos, info))
                may_set = self.get_may_set_predicate(receiver, python_field, ctx, pos)
                result.append(self.viper.Inhale(may_set, pos, info))
            else:
                raise UnsupportedException(node)
        return result

    def translate_stmt_AugAssign(self, node: ast.AugAssign,
                                 ctx: Context) -> List[Stmt]:
        left_stmt, left = self.translate_expr(node.target, ctx, as_read=True)
        if left_stmt:
            raise InvalidProgramException(node, 'purity.violated')
        stmt, right = self.translate_expr(node.value, ctx)
        left_type = self.get_type(node.target, ctx)
        right_type = self.get_type(node.value, ctx)
        op_stmt, result = self.translate_operator(left, right, left_type,
                                                  right_type, node, ctx)
        stmt += op_stmt
        result = self.to_ref(result, ctx)
        assign_stmts, _ = self.assign_to(node.target, result, None, None, right_type,
                                         node, ctx)
        return stmt + assign_stmts

    def translate_stmt_Pass(self, node: ast.Pass, ctx: Context) -> List[Stmt]:
        return []

    def _create_for_loop_invariant(self, iter_var: PythonVar, seq_temp_var: PythonVar,
                                   target_var: PythonVar,
                                   err_var: PythonVar,
                                   iterable: Expr,
                                   iterable_type: PythonType,
                                   assign_expr: List[Expr],
                                   node: ast.AST,
                                   ctx: Context) -> List[Expr]:
        """
        Creates the default invariant for for loops using iterators. It's a
        static block of code that's always the same except for possible boxing
        and unboxing, and looks like this:

        .. code-block:: silver
            invariant acc(a.list_acc, 1 / 20)
            invariant acc(iter.list_acc, 1 / 20)
            invariant iter.list_acc == list___sil_seq__(a)
            invariant acc(iter.__iter_index, write)
            invariant acc(iter.__previous, write)
            invariant iter_err == null ==>
                      iter.__iter_index - 1 == |iter.__previous|
            invariant iter_err != null ==>
                      iter.__iter_index == |iter.__previous|
            invariant iter.__iter_index >= 0 &&
                      iter.__iter_index <= |iter.list_acc|
            invariant |iter.list_acc| > 0 ==>
                      c == iter.list_acc[iter.__iter_index - 1]
            invariant (if there are multiple loop targets, relations between
                       c and the loop targets, e.g. a == getitem(c, 0) etc.,
                       as given in ``assign_expr``)
            invariant |iter.list_acc| > 0 ==> (c in iter.list_acc)
            invariant iter_err == null ==>
                          iter.__previous ==
                          iter.list_acc[..iter.__iter_index - 1]
            invariant |iter.list_acc| > 0 ==>
                      issubtype(typeof(c), list()) && ...
            invariant iter_err != null ==>
                      iter.__previous == iter.list_acc
        """
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        seq_ref = self.viper.SeqType(self.viper.Ref)
        set_ref = self.viper.SetType(self.viper.Ref)

        iter_seq = self.get_sequence(iterable_type, iterable, None, node, ctx, pos)
        full_perm = self.viper.FullPerm(pos, info)

        invariant = []
        one = self.viper.IntLit(1, pos, info)
        zero = self.viper.IntLit(0, pos, info)
        twenty = self.viper.IntLit(20, pos, info)
        frac_perm_120 = self.viper.FractionalPerm(one, twenty, pos, info)
        if iterable_type.name in {DICT_TYPE, LIST_TYPE, SET_TYPE}:
            field_name = iterable_type.name + '_acc'
            field_type = seq_ref if iterable_type.name == LIST_TYPE else set_ref
            acc_field = self.viper.Field(field_name, field_type, pos, info)
            field_acc = self.viper.FieldAccess(iterable, acc_field, pos, info)
            field_pred = self.viper.FieldAccessPredicate(field_acc,
                                                         frac_perm_120, pos,
                                                         info)
            invariant.append(field_pred)
        elif iterable_type.name == RANGE_TYPE:
            pass
        else:
            raise UnsupportedException(node)

        list_acc_field = self.viper.Field('list_acc', seq_ref, pos, info)
        iter_acc = self.viper.FieldAccess(iter_var.ref(), list_acc_field, pos,
                                          info)
        iter_acc_pred = self.viper.FieldAccessPredicate(iter_acc, frac_perm_120,
                                                        pos, info)
        invariant.append(iter_acc_pred)

        iter_list_equal = self.viper.EqCmp(iter_acc, iter_seq, pos, info)
        invariant.append(iter_list_equal)

        iter_list_equal = self.viper.EqCmp(seq_temp_var.ref(), iter_seq, pos, info)
        invariant.append(iter_list_equal)

        index_field = self.viper.Field('__iter_index', self.viper.Int, pos,
                                       info)
        iter_index_acc = self.viper.FieldAccess(iter_var.ref(), index_field,
                                                pos, info)
        iter_index_acc_pred = self.viper.FieldAccessPredicate(iter_index_acc,
                                                              full_perm, pos,
                                                              info)
        invariant.append(iter_index_acc_pred)

        previous_field = self.viper.Field('__previous', seq_ref, pos, info)
        previous_list_acc = self.viper.FieldAccess(iter_var.ref(),
                                                   previous_field,
                                                   pos, info)
        previous_list_acc_pred = self.viper.FieldAccessPredicate(
            previous_list_acc, full_perm, pos, info)
        invariant.append(previous_list_acc_pred)

        index_minus_one = self.viper.Sub(iter_index_acc, one, pos, info)

        previous_len = self.viper.SeqLength(previous_list_acc, pos, info)
        no_error_previous_len_eq = self.viper.EqCmp(index_minus_one,
                                                    previous_len, pos, info)
        error_previous_len_eq = self.viper.EqCmp(iter_index_acc, previous_len,
                                                 pos, info)

        null = self.viper.NullLit(pos, info)

        no_error = self.viper.EqCmp(err_var.ref(), null, pos, info)
        some_error = self.viper.NeCmp(err_var.ref(), null, pos, info)

        invariant.append(self.viper.Implies(no_error, no_error_previous_len_eq,
                                            pos, info))
        invariant.append(self.viper.Implies(some_error, error_previous_len_eq,
                                            pos, info))

        index_nonneg = self.viper.GeCmp(iter_index_acc, zero, pos, info)
        iter_list_len = self.viper.SeqLength(iter_acc, pos, info)

        non_empty_iterator = self.viper.GtCmp(iter_list_len, zero, pos, info)
        empty_iterator = self.viper.EqCmp(iter_list_len, zero, pos, info)

        no_error_implies_non_empty = self.viper.Implies(no_error,
                                                        non_empty_iterator, pos,
                                                        info)
        invariant.append(no_error_implies_non_empty)

        index_le_len = self.viper.LeCmp(iter_index_acc, iter_list_len, pos,
                                        info)
        index_bounds = self.viper.And(index_nonneg, index_le_len, pos, info)
        invariant.append(index_bounds)

        iter_current_index = self.viper.SeqIndex(iter_acc, index_minus_one, pos,
                                                 info)
        boxed_target = target_var.ref()
        if target_var.type.name in PRIMITIVES:
            boxed_target = self.box_primitive(boxed_target, target_var.type,
                                              None, ctx)
            iter_current_index = self.unbox_primitive(iter_current_index,
                                                      target_var.type, None,
                                                      ctx)

        positive_index = self.viper.GtCmp(iter_index_acc, zero, pos, info)
        invariant.append(self.viper.Implies(non_empty_iterator, positive_index,
                                            pos, info))
        current_element_index = self.viper.EqCmp(target_var.ref(),
                                                 iter_current_index, pos, info)
        current_element_contained = self.viper.SeqContains(boxed_target,
                                                           iter_acc, pos, info)
        invariant.append(self.viper.Implies(non_empty_iterator,
                                            current_element_index, pos, info))
        invariant.append(self.viper.Implies(non_empty_iterator,
                                            current_element_contained, pos,
                                            info))

        previous_elements = self.viper.SeqTake(iter_acc, index_minus_one, pos,
                                               info)
        iter_previous_contents = self.viper.EqCmp(previous_list_acc,
                                                  previous_elements, pos, info)
        invariant.append(self.viper.Implies(no_error, iter_previous_contents,
                                            pos, info))

        target_type = self.type_check(target_var.ref(), target_var.type, pos,
                                      ctx)
        invariant.append(self.viper.Implies(non_empty_iterator, target_type,
                                            pos, info))
        # Add information about and permissions for actual loop targets
        for target_info in assign_expr:
            invariant.append(self.viper.Implies(non_empty_iterator, target_info,
                                                pos, info))

        previous_is_all = self.viper.EqCmp(previous_list_acc, iter_acc, pos,
                                           info)
        invariant.append(self.viper.Implies(some_error, previous_is_all, pos,
                                            info))
        invariant.append(self.viper.Implies(empty_iterator, some_error, pos, info))
        return invariant

    def _get_iterator(self, iterable: Expr, iterable_type: PythonType,
                      node: ast.AST, ctx: Context) -> Tuple[PythonVar,
                                                            List[Stmt]]:
        iter_class = ctx.module.global_module.classes['Iterator']
        iter_var = ctx.actual_function.create_variable('iter', iter_class,
                                                       self.translator)
        assert not node in ctx.loop_iterators
        ctx.loop_iterators[node] = iter_var
        args = [iterable]
        arg_types = [iterable_type]
        iter_assign = self.get_method_call(iterable_type, '__iter__', args,
                                           arg_types, [iter_var.ref()], node,
                                           ctx)
        return iter_var, iter_assign

    def _get_next_call(self, iter_var: PythonVar, target_var: PythonVar,
                       node: ast.For,
                       ctx: Context) -> Tuple[PythonVar, List[Stmt]]:
        exc_class = ctx.module.global_module.classes['Exception']
        err_var = ctx.actual_function.create_variable('iter_err', exc_class,
                                                      self.translator)
        iter_class = ctx.module.global_module.classes['Iterator']
        args = [iter_var.ref()]
        arg_types = [iter_class]
        targets = [target_var.ref(node.target, ctx), err_var.ref()]
        next_call = self.get_method_call(iter_class, '__next__', args,
                                         arg_types, targets, node, ctx)
        return err_var, next_call

    def _get_iterator_delete(self, iter_var: PythonVar, node: ast.For,
                             ctx: Context) -> List[Stmt]:
        iter_class = ctx.module.global_module.classes['Iterator']
        args = [iter_var.ref()]
        arg_types = [iter_class]
        iter_del = self.get_method_call(iter_class, '__del__', args, arg_types,
                                        [], node, ctx)
        return iter_del

    def _get_havocked_vars(self, nodes: List[ast.AST],
                           ctx: Context) -> List[PythonVar]:
        """
        Finds all local variables written to within the given partial ASTs which
        already existed before.
        """
        result = []
        collector = AssignCollector()
        for stmt in nodes:
            collector.visit(stmt)
        for name in collector.assigned_vars:
            if name in ctx.var_aliases:
                var = ctx.var_aliases[name]
            else:
                var = ctx.actual_function.get_variable(name)
            if (name in ctx.actual_function.args or
                    (var.writes and not contains_stmt(nodes, var.writes[0]))):
                result.append(var)
        return result

    def _get_havocked_module_var_info(self, ctx: Context) -> StmtsAndExpr:
        """
        For global loops, saves the information which names are defined.
        """
        no_pos = self.no_position(ctx)
        no_info = self.no_info(ctx)
        if not self.is_main_method(ctx):
            return [], self.viper.TrueLit(no_pos, no_info)
        set_type = SilverType(self.viper.SetType(self.name_type()), ctx.module)
        tmp_var = ctx.current_function.create_variable('current_names', set_type,
                                                       self.translator)
        assign = self.viper.LocalVarAssign(tmp_var.ref(), ctx.module.names_var[1], no_pos,
                                           no_info)
        subset = self.viper.AnySetSubset(tmp_var.ref(), ctx.module.names_var[1], no_pos,
                                         no_info)
        return [assign], subset


    def _get_havocked_var_type_info(self, nodes: List[ast.AST],
                                    ctx: Context) -> List[Expr]:
        """
        Creates a list of assertions containing type information for all local
        variables written to within the given partial ASTs which already
        existed before.
        To be used to remember type information about arguments/local variables
        which are assigned to in loops and therefore havocked.
        """
        result = []
        if self.is_main_method(ctx):
            return result
        vars = self._get_havocked_vars(nodes, ctx)
        for var in vars:
            ref = var.ref()
            result.append(self.type_check(ref, var.type,
                                          self.no_position(ctx), ctx))
        return result

    def translate_stmt_For(self, node: ast.For, ctx: Context) -> List[Stmt]:
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)
        post_label = ctx.actual_function.get_fresh_name('post_loop')
        end_label = ctx.actual_function.get_fresh_name('loop_end')
        iterable_type = self.get_type(node.iter, ctx)
        iterable_stmt, iterable = self.translate_expr(node.iter, ctx)
        iterable_var = ctx.actual_function.create_variable('iterable', iterable_type,
                                                           self.translator, True)
        iterable_assign = self.viper.LocalVarAssign(iterable_var.ref(), iterable,
                                                    position, info)
        iterable = iterable_var.ref()
        iterable_stmt.append(iterable_assign)
        iter_var, iter_assign = self._get_iterator(iterable, iterable_type,
                                                   node, ctx)
        # Find type of the collection content we're iterating over.
        if iterable_type.name in (LIST_TYPE, DICT_TYPE, SET_TYPE):
            target_type = iterable_type.type_args[0]
        elif iterable_type.name in (RANGE_TYPE, BYTES_TYPE):
            target_type = ctx.module.global_module.classes[INT_TYPE]
        else:
            raise UnsupportedException(node, 'unknown.iterable')

        # Create artificial new variable to store current iteration content.
        target_var = ctx.actual_function.create_variable('loop_target',
                                                         target_type,
                                                         self.translator)

        err_var, next_call = self._get_next_call(iter_var, target_var,
                                                 node, ctx)

        # Assign target_var contents to actual loop target(s).
        assign_stmt, assign_expr = self.assign_to(node.target,
                                                  target_var.ref(),
                                                  None, None, target_type,
                                                  node, ctx)

        cond = self.viper.EqCmp(err_var.ref(),
                                self.viper.NullLit(position, info),
                                position, info)

        conditional_assign = self.viper.If(cond, self.translate_block(assign_stmt,
                                                                      position, info),
                                           self.translate_block([], position, info),
                                           position, info)
        assign_stmt = [conditional_assign]

        seq_ref = self.viper.SeqType(self.viper.Ref)
        seq_ref_type = SilverType(seq_ref, ctx.module)

        seq_temp_var = ctx.current_function.create_variable('seqtmp', seq_ref_type,
                                                            self.translator)

        iter_seq = self.get_sequence(iterable_type, iterable, None, node, ctx, position)

        seq_temp_assign = self.viper.LocalVarAssign(seq_temp_var.ref(), iter_seq,
                                                    position, info)

        self.enter_loop_translation(node, post_label, end_label, ctx, err_var)

        invariant = self._create_for_loop_invariant(iter_var, seq_temp_var, target_var,
                                                    err_var, iterable,
                                                    iterable_type, assign_expr,
                                                    node, ctx)
        start, end = get_body_indices(node.body)

        global_stmts, global_inv = self._get_havocked_module_var_info(ctx)
        invariant.append(global_inv)
        # Remember type information about havocked local variables.
        invariant.extend(self._get_havocked_var_type_info(node.body[start:end],
                                                          ctx))


        for expr, aliases in ctx.actual_function.loop_invariants[node]:
            with ctx.additional_aliases(aliases):
                invariant.append(self.translate_contract(expr, ctx))

        body = flatten(
            [self.translate_stmt(stmt, ctx) for stmt in node.body[start:end]])
        # Label for continue to jump to
        body.append(self.viper.Label(end_label, position, info))
        body.extend(next_call)

        body.extend(assign_stmt)

        loop = global_stmts + self.create_while_node(
            ctx, cond, invariant, [], body, node)
        iter_del = self._get_iterator_delete(iter_var, node, ctx)
        self.leave_loop_translation(ctx)
        del ctx.loop_iterators[node]
        result = (iterable_stmt + iter_assign + next_call + assign_stmt +
                  [seq_temp_assign] + loop + iter_del)
        result += self._set_result_none(ctx)
        if node.orelse:
            translated_block = flatten([self.translate_stmt(stmt, ctx) for stmt
                                        in node.orelse])
            result += translated_block
        # Label for break to jump to
        result.append(self.viper.Label(post_label, position, info))
        result += self._set_result_none(ctx)
        return result

    def translate_stmt_Assert(self, node: ast.Assert,
                              ctx: Context) -> List[Stmt]:
        stmt, expr = self.translate_expr(node.test, ctx, self.viper.Bool)
        assertion = self.viper.Assert(expr, self.to_position(node, ctx),
                                      self.no_info(ctx))
        return stmt + [assertion]

    def _get_try_block(self, node: Stmt, ctx: Context) -> PythonTryBlock:
        try_block = None
        for block in ctx.actual_function.try_blocks:
            if block.node is node:
                try_block = block
                break
        return try_block

    def translate_stmt_With(self, node: ast.With, ctx: Context) -> List[Stmt]:
        try_block = self._get_try_block(node, ctx)
        assert try_block
        code_var = try_block.get_finally_var(self.translator)
        if code_var.sil_name in ctx.var_aliases:
            code_var = ctx.var_aliases[code_var.sil_name]
        code_var = code_var.ref()
        zero = self.viper.IntLit(0, self.no_position(ctx), self.no_info(ctx))
        # Get context mgr
        ctx_stmt, ctx_mgr = self.translate_expr(try_block.with_item.context_expr,
                                                ctx)
        ctx_type = self.get_type(try_block.with_item.context_expr, ctx)
        enter_method = ctx_type.get_method('__enter__')
        # Create temp var
        enter_res_type = enter_method.type
        with_ctx = ctx.current_function.create_variable('with_ctx',
                                                         ctx_type,
                                                         self.translator)
        try_block.with_var = with_ctx
        ctx_assign = self.viper.LocalVarAssign(with_ctx.ref(), ctx_mgr,
                                               self.no_position(ctx),
                                               self.no_info(ctx))
        enter_res = ctx.current_function.create_variable('enter_res',
                                                         enter_res_type,
                                                         self.translator)
        # Call enter
        enter_call = self.get_method_call(ctx_type, '__enter__',
                                          [with_ctx.ref()],
                                          [ctx_type],
                                          [enter_res.ref(node, ctx)], node, ctx)

        assign = self.viper.LocalVarAssign(code_var, zero,
                                           self.no_position(ctx),
                                           self.no_info(ctx))
        if try_block.with_item.optional_vars:
            as_expr = try_block.with_item.optional_vars
            as_var = ctx.current_function.get_variable(as_expr.id)
            enter_assign = self.viper.LocalVarAssign(as_var.ref(as_expr, ctx),
                                                     enter_res.ref(),
                                                     self.to_position(as_expr,
                                                                      ctx),
                                                     self.no_info(ctx))
            define_var = self.set_var_defined(as_var, self.no_position(ctx),
                                              self.no_info(ctx))
            body = [enter_assign, define_var, assign]
        else:
            body = [assign]
        body += flatten([self.translate_stmt(stmt, ctx) for stmt in node.body])
        finally_name = ctx.get_label_name(try_block.finally_name)
        goto = self.viper.Goto(finally_name,
                               self.to_position(node, ctx),
                               self.no_info(ctx))
        body.append(goto)
        label_name = ctx.get_label_name(try_block.post_name)
        end_label = self.viper.Label(label_name,
                                     self.to_position(node, ctx),
                                     self.no_info(ctx))
        return ctx_stmt + [ctx_assign] + enter_call + body + [end_label]

    def translate_stmt_Try(self, node: ast.Try, ctx: Context) -> List[Stmt]:
        try_block = self._get_try_block(node, ctx)
        assert try_block
        code_var = try_block.get_finally_var(self.translator)
        if code_var.sil_name in ctx.var_aliases:
            code_var = ctx.var_aliases[code_var.sil_name]
        code_var = code_var.ref()
        zero = self.viper.IntLit(0, self.no_position(ctx), self.no_info(ctx))
        assign = self.viper.LocalVarAssign(code_var, zero,
                                           self.no_position(ctx),
                                           self.no_info(ctx))
        body = [assign]
        body += flatten([self.translate_stmt(stmt, ctx) for stmt in node.body])
        try_block.handler_aliases = ctx.var_aliases.copy()
        if try_block.else_block:
            else_label = ctx.get_label_name(try_block.else_block.name)
            goto = self.viper.Goto(else_label,
                                   self.to_position(node, ctx),
                                   self.no_info(ctx))
            body.append(goto)
        elif try_block.finally_block:
            finally_name = ctx.get_label_name(try_block.finally_name)
            goto = self.viper.Goto(finally_name,
                                   self.to_position(node, ctx),
                                   self.no_info(ctx))
            body.append(goto)
        label_name = ctx.get_label_name(try_block.post_name)
        end_label = self.viper.Label(label_name,
                                     self.to_position(node, ctx),
                                     self.no_info(ctx))
        return body + [end_label]

    def _translate_stmt_raise_create(self, node: ast.Raise,
                                     error_var: 'silver.ast.LocalVarRef',
                                     ctx: Context) -> List[Stmt]:
        """
        Translate the part of raise where we create the exception.
        """
        raised = self.get_target(node.exc, ctx)
        if (not isinstance(node.exc, ast.Call) and
                isinstance(raised, PythonType)):
            # The argument of raise is a class; call constructor with no
            # arguments.
            args = []
            init = raised.get_method('__init__')
            if init:
                _, args, _ = self.translate_args(init, [], [], node.exc, ctx,
                                                 True)
            stmt, exception = self.translate_constructor_call(raised, node.exc,
                                                              args, [], ctx)
        else:
            stmt, exception = self.translate_expr(node.exc, ctx)
        position = self.to_position(node, ctx)
        if node.cause:
            cause_stmt, cause = self.translate_expr(node.cause, ctx)
            stmt += cause_stmt
        assignment = self.viper.LocalVarAssign(error_var, exception, position,
                                               self.no_info(ctx))
        return stmt + [assignment]

    def translate_stmt_Raise(self, node: ast.Raise, ctx: Context) -> List[Stmt]:
        var = self.get_error_var(node, ctx)
        create_stmts = self._translate_stmt_raise_create(node, var, ctx)
        catchers = self.create_exception_catchers(
            var, ctx.actual_function.try_blocks, node, ctx)
        return create_stmts + catchers

    def translate_stmt_Call(self, node: ast.Call, ctx: Context) -> List[Stmt]:
        stmt, expr = self.translate_Call(node, ctx)
        if expr:
            type = self.get_type(node, ctx)
            var = ctx.current_function.create_variable('expr', type,
                                                       self.translator)
            assign = self.viper.LocalVarAssign(var.ref(node, ctx), expr,
                                               self.to_position(node, ctx),
                                               self.no_info(ctx))
            stmt.append(assign)
        return stmt

    def translate_stmt_Expr(self, node: ast.Expr, ctx: Context) -> List[Stmt]:
        if isinstance(node.value, ast.Call):
            # Call translate_Call directly to preserve information that this is
            # a top-level contract statement (those aren't allowed in most places).
            stmt, val = self.translate_Call(node.value, ctx, statement=True)
            if val is not None:
                pos = self.to_position(node, ctx)
                info = self.no_info(ctx)
                res_type = self.get_type(node.value, ctx)
                res_var = ctx.current_function.create_variable('expr_res', res_type,
                                                               self.translator)
                assign = self.viper.LocalVarAssign(res_var.ref(), val, pos, info)
                stmt.append(assign)
            return stmt
        elif isinstance(node.value, (ast.Str, ast.Ellipsis)):
            # Docstring or ellipsis, just skip.
            return []
        else:
            raise UnsupportedException(node)

    def translate_stmt_If(self, node: ast.If, ctx: Context) -> List[Stmt]:
        cond_stmt, cond = self.translate_expr(node.test, ctx,
                                              target_type=self.viper.Bool)
        then_body = flatten([self.translate_stmt(stmt, ctx)
                             for stmt in node.body])
        then_block = self.translate_block(then_body,
                                          self.to_position(node, ctx),
                                          self.no_info(ctx))
        else_body = flatten([self.translate_stmt(stmt, ctx)
                             for stmt in node.orelse])
        else_block = self.translate_block(
            else_body,
            self.to_position(node, ctx), self.no_info(ctx))
        position = self.to_position(node, ctx)
        return cond_stmt + [self.viper.If(cond, then_block, else_block,
                                          position, self.no_info(ctx))]

    def assign_to(self, lhs: ast.AST, rhs: Expr, rhs_index: Optional[int],
                  rhs_end: Optional[Expr], rhs_type: PythonType,
                  node: ast.AST, ctx: Context,
                  allow_impure: bool = False) -> Tuple[List[Stmt], List[Expr]]:
        """
        Assigns the given expression ``rhs`` to the target given in ``lhs``.
        If ``rhs_index`` is set, will only assign the element of ``rhs`` at
        this index; if ``rhs_end`` is also set, will assign a list containing
        the given range of elements to ``lhs`` (assuming ``lhs`` is of type
        ast.Starred).

        In addition to assignment statements, returns a list of assertions
        which are known to hold after the assignment, to be used in loop
        invariants.
        """
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)
        if isinstance(lhs, ast.Starred):
            return self._assign_to_starred(lhs, rhs, rhs_index, rhs_end,
                                           rhs_type, node, ctx)
        if rhs_index is not None:
            rhs_lit = self.viper.IntLit(rhs_index, position, info)
            args = [rhs, rhs_lit]
            arg_types = [rhs_type, None]
            rhs = self.get_function_call(rhs_type, '__getitem__', args,
                                         arg_types, node, ctx)
            if rhs_type.name == TUPLE_TYPE and rhs_type.exact_length:
                rhs_type = rhs_type.type_args[rhs_index]
            else:
                rhs_type = rhs_type.type_args[0]
        if isinstance(lhs, ast.Tuple):
            return self._assign_to_tuple(lhs, rhs, rhs_type, node, ctx)

        return self._assign_single_value(lhs, rhs, rhs_type, node, ctx,
                                         allow_impure=allow_impure)

    def _assign_single_value(self, lhs: ast.AST, rhs: Expr, rhs_type: PythonType,
                             node: ast.AST, ctx: Context,
                             allow_impure: bool) -> Tuple[List[Stmt], List[Expr]]:
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)
        definedness_expr = self.viper.TrueLit(position, info)

        if isinstance(lhs, ast.Subscript):
            return self._assign_with_subscript(lhs, rhs, node, ctx, allow_impure)

        target = self.get_target(lhs, ctx)
        if isinstance(target, PythonType):
            # We're assigning a type alias
            return [], []
        if isinstance(target, PythonMethod):
            # We're assigning to a property, so we have to call the method representing
            # the property setter.
            assert isinstance(lhs, ast.Attribute)
            assert target.setter
            arg_stmt, arg = self.translate_expr(lhs.value, ctx)
            call = self.create_method_call_node(ctx, target.setter.sil_name, [arg, rhs],
                                                [], position, info, target.setter, lhs)
            getter_type = self.translate_type(target.type, ctx)
            self_arg = self.viper.LocalVarDecl('self', self.viper.Ref, position, info)
            getter = self.viper.FuncApp(target.sil_name, [arg], position, info,
                                        getter_type, [self_arg])
            getter_equal = self.viper.EqCmp(getter, rhs, position, info)
            return arg_stmt + call, [getter_equal]
        if isinstance(lhs, ast.Attribute):
            type = self.get_type(lhs.value, ctx)
            if isinstance(type, UnionType) and not isinstance(type, OptionalType):
                stmt, receiver = self.translate_expr(lhs.value, ctx)
                guarded_field_assign = []
                for recv_type in toposort_classes(type.get_types() - {None}):
                    assign_guard = self.var_type_check(lhs.value.id, recv_type, position,
                                                       ctx)
                    field = recv_type.get_field(lhs.attr).actual_field
                    field_access = self.viper.FieldAccess(receiver, field.sil_field,
                                                          position, info)
                    permission = self.create_new_field_permission(field_access, field,
                                                                  position, info, ctx)
                    assign_stmt = self.viper.FieldAssign(field_access, rhs, position, info)
                    block = self.translate_block([permission, assign_stmt], position, info)
                    guarded_field_assign.append((assign_guard, block))
                chained_field_assign = chain_if_stmts(guarded_field_assign, self.viper,
                                                       position, info, ctx)
                return stmt + [chained_field_assign], None
        lhs_stmt, var = self.translate_expr(lhs, ctx)
        before_assign = []
        after_assign = []
        if isinstance(target, PythonGlobalVar):
            if target.is_final:
                # For final variables, we assume that the function representing the
                # variable is equal to the RHS of the assignment. For pure values,
                # this will do nothing, since the postcondition will already say the same;
                # for impure stuff, it will connect the function to the stuff that was
                # created.
                def assignment(lhs, rhs, pos, info):
                    eq = self.viper.EqCmp(lhs, rhs, pos, info)
                    return self.viper.Inhale(eq, position, info)
            else:
                assignment = self.viper.FieldAssign
            if self.is_main_method(ctx):
                after_assign.append(self.set_global_defined(target, ctx.module, node, ctx))
        elif isinstance(lhs, ast.Name):
            assignment = self.viper.LocalVarAssign
            if self.is_local_variable(target, ctx):
                after_assign.append(self.set_var_defined(target, position, info))
                definedness_expr = self.check_var_defined(target, position, info)
        else:
            assignment = self.viper.FieldAssign
            permission_inhale = self.create_new_field_permission(var, target,
                                                                 position, info, ctx)
            before_assign.append(permission_inhale)

        assign_stmt = assignment(var, rhs, position, info)
        assign_val = self.viper.And(self.viper.EqCmp(var, rhs, position, info),
                                    definedness_expr, position, info)
        return lhs_stmt + before_assign + [assign_stmt] + after_assign, [assign_val]

    def create_new_field_permission(self, field_acc: Expr, target: PythonField,
                                    position: Position, info: Info, ctx: Context) -> Stmt:
        """
        Creates a statement that checks if the receiver of the given field access is the
        self-parameter of the current method and there is a permission to create the
        given field. If this is the case, it will exhale the permission to create the
        field, and inhale a write permission to the field instead.
        To be used for field writes in constructors.
        """
        receiver = field_acc.rcv()

        no_perm = self.viper.NoPerm(position, info)
        id_value = self._get_string_value(target.actual_field.sil_name)
        id = self.viper.IntLit(id_value, position, info)
        may_set_pred = self.viper.PredicateAccess([receiver, id], MAY_SET_PRED, position,
                                                  info)
        may_set_perm = self.viper.CurrentPerm(may_set_pred, position, info)
        may_set = self.viper.PermGtCmp(may_set_perm, no_perm, position, info)
        full_perm = self.viper.FullPerm(position, info)
        all_may_set = self.viper.PredicateAccessPredicate(may_set_pred, full_perm,
                                                          position, info)
        field_perm = self.viper.FieldAccessPredicate(field_acc, full_perm, position, info)
        exhale = self.viper.Exhale(all_may_set, position, info)
        inhale = self.viper.Inhale(field_perm, position, info)
        in_ex = self.translate_block([exhale, inhale], position, info)
        empty_block = self.translate_block([], position, info)
        inner_if = self.viper.If(may_set, in_ex, empty_block, position, info)
        return inner_if

    def _assign_with_subscript(self, lhs: ast.Tuple, rhs: Expr, node: ast.AST,
                               ctx: Context, allow_impure: bool) -> Tuple[List[Stmt],
                                                                          List[Expr]]:
        # Special treatment for subscript; instead of an assignment, we
        # need to call a setitem method.
        if not isinstance(node.targets[0].slice, ast.Index):
            raise UnsupportedException(node, 'assignment to slice')
        position = self.to_position(node, ctx)
        target_cls = self.get_type(lhs.value, ctx)
        lhs_stmt, target = self.translate_expr(lhs.value, ctx)
        ind_stmt, index = self.translate_expr(lhs.slice.value, ctx,
                                              target_type=self.viper.Int)
        args = [target, index, rhs]
        arg_types = [None, None, None]
        stmt = self.get_method_call(target_cls, '__setitem__', args,
                                    arg_types, [], node, ctx)
        # The respective assertion states that getitem with the given index
        # now has the assigned value.
        if allow_impure:
            item_stmt, item = self.get_func_or_method_call(target_cls, '__getitem__',
                                                           [target, index], [None, None],
                                                           node, ctx)
            val = None
        else:
            item = self.get_function_call(
                target_cls, '__getitem__', [target, index], [None, None], node, ctx)
            val = self.viper.EqCmp(item, rhs, position, self.no_info(ctx))
        return lhs_stmt + ind_stmt + stmt, [val]

    def _assign_to_tuple(self, lhs: ast.Tuple, rhs: Expr, rhs_type: PythonType,
                         node: ast.AST,
                         ctx: Context) -> Tuple[List[Stmt], List[Expr]]:
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)

        no_after_starred = 0
        next = 0
        stmt_result = []
        val_result = []
        # Need to find out how many other receivers come after a starred
        # expression (if any) to calculate the last index that should be
        # assigned to the starred expression.
        for index, e in enumerate(lhs.elts):
            if isinstance(e, ast.Starred):
                no_after_starred = len(lhs.elts) - index - 1
                next = -no_after_starred
            else:
                next = next + 1
            next_expr = self.viper.IntLit(next, position, info)
            if next <= 0:
                # Calculate next index dynamically as len(rhs) + next
                # because len(rhs) might be unknown statically.
                len_expr = self.get_function_call(rhs_type, '__len__',
                                                  [rhs], [None], node, ctx)
                next_expr = self.viper.Add(len_expr, next_expr, position,
                                           info)

            stmt, val = self.assign_to(e, rhs, index, next_expr,
                                       rhs_type, node, ctx)
            stmt_result += stmt
            val_result += val
        return stmt_result, val_result

    def _assign_to_starred(self, lhs: ast.Starred, rhs: Expr,
                           rhs_index: Optional[int], rhs_end: Optional[Expr],
                           rhs_type: PythonType, node: ast.AST,
                           ctx: Context) -> Tuple[List[Stmt], List[Expr]]:
        assert rhs_index is not None and rhs_end
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)

        rhs_lit = self.viper.IntLit(rhs_index, position, info)
        list_class = ctx.module.global_module.classes[LIST_TYPE]
        stmt, res_var = self.translate_expr(lhs.value, ctx)

        # Create a new list and assign it to starred variable
        constr_call = self.get_method_call(list_class, '__init__', [], [],
                                           [res_var], node, ctx)
        stmt += constr_call
        # Inhale the type of the newly created list (including type arguments)
        list_type = self.get_type(lhs.value, ctx)
        position = self.to_position(node, ctx)
        stmt.append(
            self.viper.Inhale(self.type_check(res_var,
                                              list_type, position, ctx),
                              position, info))
        # Set list contents to segment of rhs from rhs_index until rhs_end
        seq = self.get_sequence(rhs_type, rhs, None, node, ctx, position)

        seq_until = self.viper.SeqTake(seq, rhs_end, position, info)
        seq_from = self.viper.SeqDrop(seq_until, rhs_lit, position, info)
        list_field = self.viper.Field('list_acc',
                                      self.viper.SeqType(self.viper.Ref),
                                      self.no_position(ctx), info)
        list_field_acc = self.viper.FieldAccess(res_var, list_field,
                                                position, info)
        # Also return new list permission...
        val = []
        full = self.viper.FullPerm(position, info)
        list_perm = self.viper.FieldAccessPredicate(list_field_acc,
                                                    full, position,
                                                    info)
        val.append(list_perm)
        list_type = self.type_check(res_var, list_type, position, ctx)
        val.append(list_type)
        assign_stmt = self.viper.FieldAssign(list_field_acc, seq_from, position,
                                             info)
        after_assign = []
        target = self.get_target(lhs.value, ctx)
        if self.is_local_variable(target, ctx):
            after_assign.append(self.set_var_defined(target, position, info))
        # ... and information about the list contents
        assign_val = self.viper.EqCmp(list_field_acc, seq_from, position, info)
        return stmt + [assign_stmt] + after_assign, val + [assign_val]

    def translate_stmt_Assign(self, node: ast.Assign,
                              ctx: Context) -> List[Stmt]:
        if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
            if node.targets[0].id in ctx.module.type_vars:
                # this is a type var assignment
                return []
            if node.targets[0].id in ctx.module.classes:
                # this is a type alias assignment
                return []
        if is_get_ghost_output(node):
            return self.translate_get_ghost_output(node, ctx)
        rhs_type = self.get_type(node.value, ctx)
        rhs_stmt, rhs = self.translate_expr(node.value, ctx)
        assign_stmts = []
        for target in node.targets:
            target_stmt, _ = self.assign_to(target, rhs, None, None, rhs_type,
                                            node, ctx, allow_impure=True)
            assign_stmts += target_stmt
        return rhs_stmt + assign_stmts

    def _translate_while_invariants(self, node: ast.While, ctx: Context) -> List[Stmt]:
        invariants = []
        for expr, aliases in ctx.actual_function.loop_invariants[node]:
            with ctx.additional_aliases(aliases):
                invariants.append(self.translate_contract(expr, ctx))
        return invariants

    def _translate_while_body(self, node: ast.While, ctx: Context, end_label: str) -> List[Stmt]:
        start, end = get_body_indices(node.body)
        body = flatten(
            [self.translate_stmt(stmt, ctx) for stmt in node.body[start:end]])
        body.append(self.viper.Label(end_label, self.to_position(node, ctx),
                                     self.no_info(ctx)))
        return body

    def _while_postamble(self, node: ast.While, post_label: str, ctx: Context) -> List[Stmt]:
        postamble = [self.viper.Label(post_label, self.to_position(node, ctx),
                                      self.no_info(ctx))]
        postamble += self._set_result_none(ctx)
        return postamble

    def translate_stmt_While(self, node: ast.While,
                             ctx: Context) -> List[Stmt]:
        post_label = ctx.actual_function.get_fresh_name('post_loop')
        end_label = ctx.actual_function.get_fresh_name('loop_end')
        self.enter_loop_translation(node, post_label, end_label, ctx)
        cond_stmt, cond = self.translate_expr(node.test, ctx,
                                              target_type=self.viper.Bool)
        if cond_stmt:
            raise InvalidProgramException(node, 'purity.violated')
        invariants = self._translate_while_invariants(node, ctx)
        locals = []
        start, end = get_body_indices(node.body)
        var_types = self._get_havocked_var_type_info(node.body[start:end], ctx)
        global_stmts, global_inv = self._get_havocked_module_var_info(ctx)
        invariants = [global_inv] + var_types + invariants
        body = self._translate_while_body(node, ctx, end_label)
        loop = global_stmts + self.create_while_node(
            ctx, cond, invariants, locals, body, node)
        self.leave_loop_translation(ctx)
        if node.orelse:
            translated_block = flatten([self.translate_stmt(stmt, ctx) for stmt
                                        in node.orelse])
            loop += translated_block
        loop += self._while_postamble(node, post_label, ctx)
        return loop

    def enter_loop_translation(
            self, node: Union[ast.While, ast.For], post_label: str,
            end_label: str, ctx: Context,
            err_var: PythonVar = None) -> None:
        self.loops[node] = (post_label, end_label)
        super().enter_loop_translation(node, ctx, err_var)

    def leave_loop_translation(self, ctx: Context) -> None:
        super().leave_loop_translation(ctx)

    def _set_result_none(self, ctx: Context) -> List[Stmt]:
        """
        Sets the return variable of the current function to null (the default
        return variable), to be used after loops which may havoc the result
        variable (if there is a return within the loop body).
        """
        result = []
        null = self.viper.NullLit(self.no_position(ctx), self.no_info(ctx))
        if ctx.actual_function.type:
            result_none = self.viper.LocalVarAssign(
                ctx.actual_function.result.ref(),
                null, self.no_position(ctx),
                self.no_info(ctx))
            result.append(result_none)
        # Do the same for the error variable
        if ctx.actual_function.declared_exceptions:
            error_none = self.viper.LocalVarAssign(
                ctx.actual_function.error_var.ref(),
                null, self.no_position(ctx), self.no_info(ctx))
            result.append(error_none)
        return result

    def translate_stmt_Break(self, node: ast.Break, ctx: Context) -> List[Stmt]:
        loop = get_parent_of_type(node, (ast.While, ast.For))
        loop_and_label = self.loops[loop]
        result = self.viper.Goto(loop_and_label[0], self.to_position(node, ctx),
                                 self.no_info(ctx))
        return [result]

    def translate_stmt_Continue(self, node: ast.Continue,
                                ctx: Context) -> List[Stmt]:

        parent = node
        # Find the loop surrounding this node.
        while not isinstance(parent._parent, (ast.While, ast.For)):
            # If we find, on the way, that we're in a try block, namely in the finally branch
            if isinstance(parent._parent, ast.Try) and parent in parent._parent.finalbody:
                # this is illegal in Python any mypy doesn't check it.
                raise InvalidProgramException(node, 'continue.in.finally')
            else:
                parent = parent._parent

        loop = parent._parent
        loop_and_label = self.loops[loop]
        result = self.viper.Goto(loop_and_label[1], self.to_position(node, ctx),
                                 self.no_info(ctx))
        return [result]

    def _translate_return(self, node: ast.Return, ctx: Context) -> List[Stmt]:
        if not node.value:
            return []
        type_ = ctx.actual_function.type
        rhs_stmt, rhs = self.translate_expr(node.value, ctx)
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        if not ctx.result_var:
            null = self.viper.NullLit(pos, info)
            assert_pos = self.to_position(node, ctx, error_string="Valid return")
            assign = self.viper.Assert(self.viper.EqCmp(rhs, null, assert_pos, info),
                                       assert_pos, info)
        else:
            assign = self.viper.LocalVarAssign(
                ctx.result_var.ref(node, ctx),
                rhs, pos, info)

        return rhs_stmt + [assign]

    def translate_stmt_Return(self, node: ast.Return,
                              ctx: Context) -> List[Stmt]:
        return_stmts = self._translate_return(node, ctx)
        tries = get_surrounding_try_blocks(ctx.actual_function.try_blocks,
                                           node)
        for try_block in tries:
            if try_block.finally_block or try_block.with_item:
                lhs = try_block.get_finally_var(self.translator).ref()
                rhs = self.viper.IntLit(1, self.no_position(ctx),
                                        self.no_info(ctx))
                finally_assign = self.viper.LocalVarAssign(lhs, rhs,
                    self.no_position(ctx), self.no_info(ctx))
                label_name = ctx.get_label_name(try_block.finally_name)
                jmp = self.viper.Goto(label_name,
                                      self.to_position(node, ctx),
                                      self.no_info(ctx))
                return return_stmts + [finally_assign, jmp]
        end_label = ctx.get_label_name(END_LABEL)
        jmp_to_end = self.viper.Goto(end_label, self.to_position(node, ctx),
                                     self.no_info(ctx))
        return return_stmts + [jmp_to_end]



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Nagini tests based on pytest framework.

Nagini tests are based on ideas taken from ``Silver``. Each test is a
Python source file with annotations that specify the expected behaviour.
The goal of the test suite is to catch changes in the behaviour,
therefore, annotations must be always up-to-date. Annotations are
written in Python comments that start with ``::``. Multiple annotations
on the same line are separated by ``|``.

Supported annotation types are:

1.  ``ExpectedOutput<(backend)>(<error_id>, <via1>, <via2>,…)`` –
    indicates that the following line should produce the specified
    error.
2.  ``UnexpectedOutput<(backend)>(<error_id>, <issue>, <via1>, <via2>,…)``
    – indicates that the following line should not produce the
    specified error, but it currently does. The problem is currently
    tracked in ``backend`` (if missing, then Nagini) issue tracker's
    issue ``issue``.
3.  ``MissingOutput<(backend)>(<error_id>, <issue>, <via1>, <via2>,…)`` –
    indicates that the error mentioned in the matching
    ``ExpectedOutput`` annotation is not produced due to issue
    ``issue``.
4.  ``Label(via)`` – mark location to be used in other annotations.
5.  ``IgnoreFile(<issue>)`` – mark that file cannot be tested due to
    critical issue such as a crash, which is tracked in ``issue``.
"""


import abc
import os
import pytest
import re
import tokenize
from collections import Counter
from typing import Any, Dict, List, Optional


# These imports monkey-patch mypy and should happen as early as possible.
import nagini_translation.mypy_patches.column_info_patch
import nagini_translation.mypy_patches.optional_patch

from nagini_translation.lib import config, jvmaccess
from nagini_translation.lib.errors import error_manager
from nagini_translation.lib.typeinfo import TypeException
from nagini_translation.lib.util import InvalidProgramException
from nagini_translation.main import translate, verify, TYPE_ERROR_PATTERN
from nagini_translation.verifier import VerificationResult, ViperVerifier


os.environ['MYPYPATH'] = config.mypy_path

_JVM = jvmaccess.JVM(config.classpath)


_MYPY_ERROR_MATCHER = re.compile(TYPE_ERROR_PATTERN)

_BACKEND_SILICON = 'silicon'
_BACKEND_CARBON = 'carbon'
_BACKEND_ANY = 'ANY'


def _consume(key: str, dictionary: Dict[str, Any], check: bool = False) -> Any:
    """Destructive read from the dictionary.

    If ``check`` is ``True``, check that read value does not evaluate to
    ``false``.
    """
    value = dictionary[key]
    del dictionary[key]
    if check:
        assert value, "{} in {} is False".format(key, dictionary)
    return value


def _consume_list(key: str, dictionary: Dict[str, Any]) -> Any:
    """Destructive read of comma separated list from the dictionary."""
    value = _consume(key, dictionary)
    if value:
        return [part for part in value.split(', ') if part]
    else:
        return []


class Error(abc.ABC):
    """Base class for reported errors.

    Subclasses of this class are wrappers that unify interfaces of three
    error types currently produced by Nagini:

    1.  Type errors produced by Mypy.
    2.  Invalid program errors produced by translators.
    3.  Verification errors produced by back-end verifiers.
    """

    @property
    @abc.abstractmethod
    def full_id(self) -> str:
        """Return full error id."""

    @property
    @abc.abstractmethod
    def line(self) -> int:
        """Return line number."""

    @abc.abstractmethod
    def get_vias(self) -> List[int]:
        """Return a list of vias."""


class VerificationError(Error):
    """Verification error reported by verifier."""

    def __init__(self, actual_error: 'Error') -> None:
        self._error = actual_error

    def __repr__(self) -> str:
        return 'VerificationError({}, line={}, vias={})'.format(
            self.full_id, self.line, self.get_vias())

    @property
    def full_id(self) -> str:
        return self._error.full_id

    @property
    def line(self) -> int:
        return self._error.position.line

    def get_vias(self) -> List[int]:
        reason_pos = self._error.reason.position
        if reason_pos.node_id:
            vias = error_manager.get_vias(reason_pos.node_id)
            if vias:
                return [via[1].line() for via in vias]
        error_pos = self._error.position
        if error_pos.node_id:
            vias = error_manager.get_vias(error_pos.node_id)
            return [via[1].line() for via in vias]
        return []


class InvalidProgramError(Error):
    """Invalid program error reported by translator."""

    def __init__(self, exception: InvalidProgramException) -> None:
        self._exception = exception

    def __repr__(self) -> str:
        return 'InvalidProgramError({}, line={}, vias={})'.format(
            self.full_id, self.line, self.get_vias())

    @property
    def full_id(self) -> str:
        return 'invalid.program:' + self._exception.code

    @property
    def line(self) -> int:
        return self._exception.node.lineno

    def get_vias(self) -> List[int]:
        return []


class TypeCheckError(Error):
    """Type error reported by Mypy."""

    def __init__(self, msg: str) -> None:
        self._msg = msg
        match = _MYPY_ERROR_MATCHER.match(msg)
        self._groups = match.groupdict()

    def __repr__(self) -> str:
        return 'TypeCheckError({}, line={}, vias={})'.format(
            self.full_id, self.line, self.get_vias())

    @property
    def full_id(self) -> str:
        return 'type.error:' + self._groups['msg']

    @property
    def line(self) -> int:
        return int(self._groups['line'])

    def get_vias(self) -> List[int]:
        return []


class Annotation:
    """Base class for all test annotations."""

    def __init__(
            self, token: tokenize.TokenInfo,
            group_dict: Dict[str, Optional[str]]) -> None:
        self._token = token
        for key, value in group_dict.items():
            if key == 'type':
                continue
            if value:
                setter_name = '_set_' + key
                # Here we check that provided annotation does not have
                # too much stuff.
                assert hasattr(self, setter_name), (
                    "Unsupported {} for {}".format(value, self))
                getattr(self, setter_name)(value)

    @property
    def line(self) -> int:
        """Get line number of this annotation."""
        return self._token.start[0] + 1

    @property
    @abc.abstractmethod
    def backend(self) -> str:
        """Back-end which this annotation is targeting."""


class BackendSpecificAnnotationMixIn:
    """Annotation that depends on the back-end.

    The subclass is expected to define a field ``_backend``.
    """

    @property
    def backend(self) -> str:
        """Back-end which this annotation is targeting."""
        return self._backend or _BACKEND_ANY


class ErrorMatchingAnnotationMixIn:
    """An annotation that can match an error.

    The subclass is expected to define fields ``_id`` and ``_labels``.
    """

    def match(self, error: Error) -> bool:
        """Check is error matches this annotation."""
        return (self._id == error.full_id and
                self.line == error.line and
                self.get_vias() == error.get_vias())


class UsingLabelsAnnotationMixIn:
    """An annotation that can refer to labels.

    The subclass is expected to define the field ``_labels``.
    """

    def resolve_labels(
            self, labels_dict: Dict[str, 'LabelAnnotation']) -> None:
        """Resolve label names to label objects."""
        for i, label in enumerate(self._labels):
            self._labels[i] = labels_dict[label]

    def get_vias(self) -> List[int]:
        """Return vias extracted from label positions."""
        return [label.line for label in self._labels]


class ExpectedOutputAnnotation(
        BackendSpecificAnnotationMixIn,
        ErrorMatchingAnnotationMixIn,
        UsingLabelsAnnotationMixIn,
        Annotation):
    """ExpectedOutput annotation."""

    def __init__(
            self, token: tokenize.TokenInfo,
            group_dict: Dict[str, Optional[str]]) -> None:
        """ExpectedOutput constructor.

        Supported info:

        +   id – mandatory.
        +   backend – optional.
        +   labels – optional.
        """
        self._id = _consume('id', group_dict, True)
        self._backend = _consume('backend', group_dict)
        self._labels = _consume_list('labels', group_dict)
        super().__init__(token, group_dict)

    def __repr__(self) -> str:
        return 'ExpectedOutput({}, line={}, vias={})'.format(
            self._id, self.line, self.get_vias())

    @property
    def full_id(self) -> str:
        """Return full error id."""
        return self._id


class UnexpectedOutputAnnotation(
        BackendSpecificAnnotationMixIn,
        ErrorMatchingAnnotationMixIn,
        UsingLabelsAnnotationMixIn,
        Annotation):
    """UnexpectedOutput annotation."""

    def __init__(
            self, token: tokenize.TokenInfo,
            group_dict: Dict[str, Optional[str]]) -> None:
        """UnexpectedOutput constructor.

        Supported info:

        +   id – mandatory.
        +   backend – optional, ``None`` means ``nagini``.
        +   issue_id – mandatory.
        +   labels – optional.
        """
        self._id = _consume('id', group_dict, True)
        self._backend = _consume('backend', group_dict)
        self._issue_id = _consume('issue_id', group_dict, True)
        self._labels = _consume_list('labels', group_dict)
        super().__init__(token, group_dict)

    def __repr__(self) -> str:
        return 'UnexpectedOutput({}, line={}, vias={})'.format(
            self._id, self.line, self.get_vias())


class MissingOutputAnnotation(
        BackendSpecificAnnotationMixIn,
        UsingLabelsAnnotationMixIn,
        Annotation):
    """MissingOutput annotation."""

    def __init__(
            self, token: tokenize.TokenInfo,
            group_dict: Dict[str, Optional[str]]) -> None:
        """MissingOutput constructor.

        Supported info:

        +   id – mandatory.
        +   backend – optional, ``None`` means ``nagini``.
        +   issue_id – mandatory.
        +   labels – optional.
        """
        self._id = _consume('id', group_dict, True)
        self._backend = _consume('backend', group_dict)
        self._issue_id = _consume('issue_id', group_dict, True)
        self._labels = _consume_list('labels', group_dict)
        super().__init__(token, group_dict)

    def match(self, expected: ExpectedOutputAnnotation) -> bool:
        """Check if this annotation matches the given ``ExpectedOutput``.

        ``MissingOutput`` annotation indicates that the output mentioned
        in a certain ``ExpectedOutput`` annotation is not going to be
        produced due to some issue. In other words, a ``MissingOutput``
        annotation silences a matching ``ExpectedOutput`` annotation.
        Intuitively, a ``MissingOutput`` annotation matches an
        ``ExpectedOutput`` annotation if they are on the same line and
        have the same arguments.
        """
        return (self.line == expected.line and
                self._id == expected.full_id and
                self.get_vias() == expected.get_vias() and
                (self.backend == expected.backend or
                 self.backend == _BACKEND_ANY or
                 expected.backend == _BACKEND_ANY))


class LabelAnnotation(Annotation):
    """Label annotation."""

    def __init__(
            self, token: tokenize.TokenInfo,
            group_dict: Dict[str, Optional[str]]) -> None:
        """Label constructor.

        Supported info:

        +   id – mandatory.
        """
        self._id = _consume('id', group_dict, True)
        super().__init__(token, group_dict)

    @property
    def name(self) -> str:
        """Return the labels name."""
        return self._id

    @property
    def backend(self) -> str:
        """Back-end which this annotation is targeting."""
        return _BACKEND_ANY


class IgnoreFileAnnotation(
        BackendSpecificAnnotationMixIn,
        Annotation):
    """IgnoreFile annotation."""

    def __init__(
            self, token: tokenize.TokenInfo,
            group_dict: Dict[str, Optional[str]]) -> None:
        """IgnoreFile constructor.

        Supported info:

        +   id – mandatory, used as issue_id.
        +   backend – optional, ``None`` means ``nagini``.
        """
        self._issue_id = _consume('id', group_dict, True)
        assert self._issue_id.isnumeric(), "Issue id must be a number."
        self._backend = _consume('backend', group_dict)
        super().__init__(token, group_dict)


class AnnotationManager:
    """A class for managing annotations in the specific test file."""

    def __init__(self, backend: str) -> None:
        self._matcher = re.compile(
            # Annotation type such as ExpectedOutput.
            r'(?P<type>[a-zA-Z]+)'
            # To which back-end the annotation is dedicated. None means
            # both.
            r'(\((?P<backend>[a-z]+)\))?'
            r'\('
            # Error message, or label id. Matches everything except
            # comma.
            r'(?P<id>[a-zA-Z\.\(\)_\-:;\d ?\'"]+)'
            # Issue id in the issue tracker.
            r'(, (?P<issue_id>\d+))?'
            # Labels. Note that label must start with a letter.
            r'(?P<labels>(, [a-zA-Z][a-zA-Z\d_]+)+)?'
            r'\)'
        )
        self._annotations = {
            'ExpectedOutput': [],
            'UnexpectedOutput': [],
            'MissingOutput': [],
            'Label': [],
            'IgnoreFile': [],
        }
        self._backend = backend

    def _create_annotation(
            self, annotation_string: str, token: tokenize.TokenInfo) -> None:
        """Create annotation object from the ``annotation_string``."""
        match = self._matcher.match(annotation_string)
        assert match, "Failed to match: {}".format(annotation_string)
        group_dict = match.groupdict()
        annotation_type = group_dict['type']
        if annotation_type == 'ExpectedOutput':
            annotation = ExpectedOutputAnnotation(token, group_dict)
        elif annotation_type == 'UnexpectedOutput':
            annotation = UnexpectedOutputAnnotation(token, group_dict)
        elif annotation_type == 'MissingOutput':
            annotation = MissingOutputAnnotation(token, group_dict)
        elif annotation_type == 'Label':
            annotation = LabelAnnotation(token, group_dict)
        elif annotation_type == 'IgnoreFile':
            annotation = IgnoreFileAnnotation(token, group_dict)
        else:
            assert False, "Unknown annotation type: {}".format(annotation_type)
        assert annotation_type in self._annotations
        if annotation.backend in (self._backend, _BACKEND_ANY):
            self._annotations[annotation_type].append(annotation)

    def _get_expected_output(self) -> List[ExpectedOutputAnnotation]:
        """Return a final list of expected output annotations."""
        expected_annotations = self._annotations['ExpectedOutput']
        missing_annotations = set(self._annotations['MissingOutput'])
        # Filter out annotations that should be missing.
        annotations = []
        for expected in expected_annotations:
            for missing in missing_annotations:
                if missing.match(expected):
                    missing_annotations.remove(missing)
                    break
            else:
                annotations.append(expected)
        assert not missing_annotations
        # Append unexpected annotations.
        annotations.extend(self._annotations['UnexpectedOutput'])
        return annotations

    def resolve_references(self) -> None:
        """Resolve references to labels."""
        labels_dict = dict(
            (label.name, label)
            for label in self._annotations['Label'])
        for annotation_type in ['ExpectedOutput', 'UnexpectedOutput',
                                'MissingOutput']:
            for annotation in self._annotations[annotation_type]:
                annotation.resolve_labels(labels_dict)

    def check_errors(self, actual_errors: List[Error]) -> None:
        """Check if actual errors match annotations."""
        annotations = set(self._get_expected_output())
        unexpected_errors = []
        for error in actual_errors:
            for annotation in annotations:
                if annotation.match(error):
                    annotations.remove(annotation)
                    break
            else:
                unexpected_errors.append(error)
        assert not unexpected_errors
        assert not annotations

    def has_unexpected_missing(self) -> bool:
        """Check if there are unexpected or missing output annotations."""
        return (self._annotations['UnexpectedOutput'] or
                self._annotations['MissingOutput'])

    def extract_annotations(self, token: tokenize.TokenInfo) -> None:
        """Extract annotations mentioned in the token."""
        content = token.string.strip()[3:]
        stripped_list = [part.strip() for part in content.split('|')]
        for part in stripped_list:
            self._create_annotation(part, token)

    def ignore_file(self) -> bool:
        """Check if file should be ignored."""
        return bool(self._annotations['IgnoreFile'])


class AnnotatedTest:
    """A class representing an annotated test.

    An annotated test is a Python source file with annotations that
    indicate expected verification errors.
    """

    def _is_annotation(self, token: tokenize.TokenInfo) -> bool:
        """Check if token is a test annotation.

        A test annotation is a comment starting with ``#::``.
        """
        return (token.type is tokenize.COMMENT and
                token.string.strip().startswith('#:: ') and
                token.string.strip().endswith(')'))

    def get_annotation_manager(
            self, path: str, backend: str) -> AnnotationManager:
        """Create ``AnnotationManager`` for given Python source file."""
        manager = AnnotationManager(backend)
        with open(path, 'rb') as fp:
            for token in tokenize.tokenize(fp.readline):
                if self._is_annotation(token):
                    manager.extract_annotations(token)
        manager.resolve_references()
        return manager


class VerificationTest(AnnotatedTest):
    """Test for testing verification of successfully translated programs."""

    def test_file(
            self, path: str, jvm: jvmaccess.JVM, verifier: ViperVerifier,
            sif: bool, reload_resources: bool, arp: bool):
        """Test specific Python file."""
        annotation_manager = self.get_annotation_manager(path, verifier.name)
        if annotation_manager.ignore_file():
            pytest.skip('Ignored')
        path = os.path.abspath(path)
        prog = translate(path, jvm, sif=sif, arp=arp, reload_resources=reload_resources)
        assert prog is not None
        vresult = verify(prog, path, jvm, verifier, arp=arp)
        self._evaluate_result(vresult, annotation_manager, jvm, sif)

    def _evaluate_result(
            self, vresult: VerificationResult,
            annotation_manager: AnnotationManager, jvm: jvmaccess.JVM, sif: bool = False):
        """Evaluate verification result with regard to test annotations."""
        if vresult:
            actual_errors = []
        else:
            assert all(
                isinstance(error.pos(), jvm.viper.silver.ast.HasLineColumn)
                for error in vresult.errors)
            actual_errors = [
                VerificationError(error) for error in vresult.errors]
            if sif:
                # carbon will report all functional errors twice, as we model two
                # executions, therefore we filter duplicated errors here.
                # (Note: we don't make errors unique, just remove one duplicate)
                distinct = []
                reprs = map(lambda e: e.__repr__(), actual_errors)
                repr_counts = Counter(reprs)
                repr_counts = dict(map(lambda rc: (rc[0], -(-rc[1] // 2)),
                                       repr_counts.items()))
                for err in actual_errors:
                    if repr_counts[err.__repr__()] > 0:
                        distinct.append(err)
                        repr_counts[err.__repr__()] -= 1
                actual_errors = distinct
        annotation_manager.check_errors(actual_errors)
        if annotation_manager.has_unexpected_missing():
            pytest.skip('Unexpected or missing output')


_VERIFICATION_TESTER = VerificationTest()


def test_verification(path, verifier, sif, reload_resources, arp):
    """Execute provided verification test."""
    _VERIFICATION_TESTER.test_file(path, _JVM, verifier, sif, reload_resources, arp)


class TranslationTest(AnnotatedTest):
    """Test for testing translation errors."""

    def test_file(self, path: str, jvm: jvmaccess.JVM, sif: bool,
                  reload_resources: bool, arp: bool):
        """Test specific Python file."""
        annotation_manager = self.get_annotation_manager(path, _BACKEND_ANY)
        if annotation_manager.ignore_file():
            pytest.skip('Ignored')
        path = os.path.abspath(path)
        try:
            translate(path, jvm, sif=sif, arp=arp, reload_resources=reload_resources)
            actual_errors = []
        except InvalidProgramException as exp1:
            actual_errors = [InvalidProgramError(exp1)]
        except TypeException as exp2:
            actual_errors = [
                TypeCheckError(msg) for msg in exp2.messages
                if _MYPY_ERROR_MATCHER.match(msg)]
        annotation_manager.check_errors(actual_errors)
        if annotation_manager.has_unexpected_missing():
            pytest.skip('Unexpected or missing output')


_TRANSLATION_TESTER = TranslationTest()


def test_translation(path, sif, reload_resources, arp):
    """Execute provided translation test."""
    _TRANSLATION_TESTER.test_file(path, _JVM, sif, reload_resources, arp)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import jpype


class JVM:
    """
    Encapsulates access to a JVM
    """

    def __init__(self, classpath: str):
        jpype.startJVM(jpype.getDefaultJVMPath(),
                       '-Djava.class.path=' + classpath, '-Xss32m')
        self.java = jpype.JPackage('java')
        self.scala = jpype.JPackage('scala')
        self.viper = jpype.JPackage('viper')
        self.fastparse = jpype.JPackage('fastparse')

    def get_proxy(self, supertype, instance):
        return jpype.JProxy(supertype, inst=instance)

    def is_known_class(self, class_object) -> bool:
        return not isinstance(class_object, jpype.JPackage)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Wrappers for Scala error objects."""


from typing import Any, List

from nagini_translation.lib.errors.messages import ERRORS, REASONS, VAGUE_REASONS
from nagini_translation.lib.errors.rules import Rules


class Position:
    """Wrapper around ``AbstractSourcePosition``."""

    def __init__(self, position: 'ast.AbstractSourcePosition') -> None:
        self._position = position
        if hasattr(position, 'id'):
            self.node_id = position.id()
        else:
            self.node_id = None

    @property
    def file_name(self) -> str:
        """Return ``file``."""
        return self._position.file().toString()

    @property
    def line(self) -> int:
        """Return ``start.line``."""
        return self._position.line()

    @property
    def column(self) -> int:
        """Return ``start.column``."""
        return self._position.column()

    def __str__(self) -> str:
        return str(self._position)


class Reason:
    """Wrapper around ``AbstractErrorReason``."""

    def __init__(self, reason_id: str, reason: 'AbstractErrorReason',
                 node: 'ast.Node' = None, vias: List[Any] = None,
                 reason_string: str = None) -> None:
        self._reason = reason
        self._node = node
        self._reason_string = reason_string
        self.identifier = reason_id
        self.vias = vias
        self.offending_node = reason.offendingNode()
        self.position = Position(self.offending_node.pos())

    def __str__(self) -> str:
        return self.string(False)

    def string(self, show_viper_reason: bool) -> str:
        """
        Creates a string representation of this reason including a reference to the Python
        AST node that caused it.
        If no such node is available, either returns a partial message that describes the
        kind of error in general, or outputs the concrete Viper-level description of the
        error, depending on the parameter ``show_viper_reason``.
        """
        reason = self._reason_string or self._node
        if reason is None and self.identifier in VAGUE_REASONS:
            if not show_viper_reason:
                return VAGUE_REASONS[self.identifier]
            else:
                return self._reason.readableMessage()
        return REASONS[self.identifier](reason)


class Error:
    """Wrapper around ``AbstractVerificationError``."""

    def __init__(self, error: 'AbstractVerificationError', rules: Rules,
                 reason_item: Any, node: 'ast.Node' = None,
                 vias: List[Any] = None) -> None:

        # Translate error id.
        viper_reason = error.reason()
        error_id = error.id()
        reason_id = viper_reason.id()
        key = error_id, reason_id
        if key in rules:
            error_id, reason_id = rules[key]

        # Construct object.
        self._error = error
        self._node = node
        self._vias = vias
        self.identifier = error_id
        if reason_item:
            self.reason = Reason(
                reason_id, viper_reason, reason_item.node,
                reason_item.vias, reason_item.reason_string)
        else:
            self.reason = Reason(reason_id, viper_reason)
        self.position = Position(error.pos())

    def pos(self) -> 'ast.AbstractSourcePosition':
        """Get position.

        .. todo:: Marco

            This method is only for compatibility with current testing
            infrastructure and should be removed.
        """
        return self._error.pos()

    @property
    def full_id(self) -> str:
        """Full error identifier."""
        return '{}:{}'.format(self.identifier, self.reason.identifier)

    @property
    def offending_node(self) -> 'ast.Node':
        """AST node where the error occurred."""
        return self._error.offendingNode()

    @property
    def readable_message(self) -> str:
        """Readable error message."""
        return self._error.readableMessage()

    @property
    def position_string(self) -> str:
        """Full error position as a string."""
        vias = self.reason.vias or self._vias or []
        vias_string = ''.join(
            ', via {0} at {1}'.format(reason, pos)
            for reason, pos in vias)
        return '{}{}'.format(self.position, vias_string)

    @property
    def message(self) -> str:
        """Human readable error message."""
        if self._node:
            return ERRORS[self.identifier](self._node)
        # If we don't have a node, fall back to the original Silver message,
        # it's better than nothing.
        return self._error.text()

    def __str__(self) -> str:
        return self.string(False, False)

    def string(self, ide_mode: bool, show_viper_errors: bool) -> str:
        """
        Format error.

        Creates an appropriate error message (referring to the
        responsible Python code) for the given Viper error.

        The error format is either optimized for human readability or uses the same format
        as IDE-mode Viper error messages, depending on the first parameter.
        The second parameter determines if the message may show Viper-level error
        explanations if no Python-level explanation is available.
        """
        if ide_mode:
            return '{0}:{1}:{2}: error: {3} {4}'.format(
                self.position.file_name,
                self.position.line, self.position.column, self.message,
                self.reason)
        else:
            return '{0} {1} ({2})'.format(
                self.message, self.reason.string(show_viper_errors), self.position_string)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

from typing import Optional

from nagini_translation.lib.context import Context
from nagini_translation.lib.program_nodes import MethodType
from nagini_translation.lib.typedefs import DomainFuncApp
from nagini_translation.translators.type_domain_factory import TypeDomainFactory

def configure_mpp_transformation(jvm, ctrl_opt: bool, seq_opt: bool,
                                 act_opt: bool, func_opt: bool) -> None:
    """
    Configure which optimizations to apply in MPP transformation.
    - ctrl_opt: only generate those control variables which are needed.
    - seq_opt:  bunch together statements which are executed under the same condition without
                interference with the other execution.
    - act_opt:  at the beginning of each method add an 'assume p1' statement.
    - func_opt: only apply the _checkDefined and _isDefined functions in the first execution.
    """
    jvm.viper.silver.sif.SIFExtendedTransformer.optimizeControlFlow(ctrl_opt)
    jvm.viper.silver.sif.SIFExtendedTransformer.optimizeSequential(seq_opt)
    jvm.viper.silver.sif.SIFExtendedTransformer.optimizeRestrictActVars(act_opt)
    if func_opt:
        jvm.viper.silver.sif.SIFExtendedTransformer.addPrimedFuncAppReplacement(
            "_checkDefined", "first_arg")
        jvm.viper.silver.sif.SIFExtendedTransformer.addPrimedFuncAppReplacement(
            "_isDefined", "true")
    else:
        jvm.viper.silver.sif.SIFExtendedTransformer.clearPrimedFuncAppReplacement()

def _to_scala_set(jvm, inset: set):
    seq = jvm.scala.collection.mutable.ArraySeq(len(inset))
    for i, elem in enumerate(inset):
        seq.update(i, elem)
    return seq.toSet()

def set_all_low_methods(jvm, names: set) -> None:
    scala_set = _to_scala_set(jvm, names)
    jvm.viper.silver.sif.SIFExtendedTransformer.setAllLowMethods(scala_set)

def set_preserves_low_methods(jvm, names: set) -> None:
    scala_set = _to_scala_set(jvm, names)
    jvm.viper.silver.sif.SIFExtendedTransformer.setPreservesLowMethods(scala_set)

def set_equality_comp_functions(jvm, names: set) -> None:
    if not names:
        return
    names_seq = jvm.scala.collection.mutable.ArraySeq(len(names))
    for i, elem in enumerate(names):
        names_seq.update(i, (elem, elem))
    hash_map = names_seq.toMap()
    jvm.viper.silver.sif.SIFExtendedTransformer.equalityCompFunctions = hash_map

def in_postcondition_of_dyn_bound_call(type_factory: TypeDomainFactory,
                                       ctx: Context) -> Optional[DomainFuncApp]:
    """
    Determine if we are in a postcondition of a dynamically bound method. If so return the
    function application representing the type of self. Else return None.
    """
    if (ctx.current_class and
            ctx.current_function.method_type == MethodType.normal and
            ctx.obligation_context.is_translating_posts):
        return type_factory.typeof(
            next(iter(ctx.actual_function.args.values())).ref(), ctx)
    return None

def in_override_check(ctx: Context) -> bool:
    exceptional_positions = ['inheritance', 'override']
    return any(map(lambda tuple: tuple[0] in exceptional_positions, ctx.position))



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Translation of IO operation use."""


import ast

from typing import cast, List

from nagini_translation.lib.constants import CALLABLE_TYPE
from nagini_translation.lib.context import Context
from nagini_translation.lib.program_nodes import (
    PythonIOExistentialVar,
    PythonMethod,
    PythonIOOperation,
    TypeVar,
)
from nagini_translation.lib.typedefs import (
    Expr,
    Stmt,
    StmtsAndExpr,
)
from nagini_translation.lib.util import (
    get_func_name,
    InvalidProgramException,
    UnsupportedException,
)
from nagini_translation.translators.io_operation.common import (
    IOOperationCommonTranslator,
)
from nagini_translation.translators.io_operation.opener import Opener
from nagini_translation.translators.io_operation.result_translator import (
    ResultTranslator,
)
from nagini_translation.translators.io_operation.utils import (
    get_variable,
    is_top_level_assertion,
    raise_invalid_operation_use,
    raise_invalid_existential_var,
    raise_invalid_get_ghost_output,
)


class IOOperationUseTranslator(IOOperationCommonTranslator):
    """Class responsible for translating IO operation use."""

    def translate_io_operation_call(self, node: ast.Call,
                                    ctx: Context) -> StmtsAndExpr:
        """Translate a call to an IO operation.

        That is:

        1.  Emit a predicate access corresponding to the operation.
        2.  Either define getter invocations corresponding to the
            operation results, or emit equalities between each result
            and already defined getter invocation.
        """
        assert ctx.actual_function

        operation = self.get_target(node, ctx)
        parameters = operation.get_parameters()
        args = self.translate_args(node.args[:len(parameters)], parameters, ctx)
        perm = self._construct_full_perm(node, ctx)

        # Translate predicate.
        predicate = self.create_predicate_access(
            operation.sil_name, args, perm, node, ctx)

        # Translate results.
        equations = self._translate_results(operation, node, ctx)

        # And everything.
        expr = predicate
        for equation in equations:
            expr = self.viper.And(expr, equation,
                                  self.to_position(node, ctx),
                                  self.no_info(ctx))
        return [], expr

    def is_io_existential_defining_equality(self, node: ast.expr,
                                            ctx: Context) -> bool:
        """Check if ``node`` defines IO existential variable.

        That is, node is equality of form:
        ``existential_variable == something``.
        """
        if (is_top_level_assertion(node) and
                isinstance(node, ast.Compare)):
            if (len(node.ops) == 1 and
                    len(node.comparators) == 1 and
                    isinstance(node.left, ast.Name) and
                    isinstance(node.ops[0], ast.Eq)):
                var = get_variable(node.left.id, ctx)
                return (
                    isinstance(var, PythonIOExistentialVar) and
                    not var.is_defined())
        return False

    def define_io_existential(self, node: ast.Compare, ctx: Context) -> None:
        """From defining equality defines IO existential variable."""
        assert self.is_io_existential_defining_equality(node, ctx)

        # TODO: The result of this call must not only be an expression,
        # but a pure expression.
        right_stmt, right = self.translate_expr(
            node.comparators[0], ctx)
        assert not right_stmt   # Should be handled by expression=True.

        name_node = cast(ast.Name, node.left)
        var = get_variable(name_node.id, ctx)

        expression_type = self.get_type(node.comparators[0], ctx)
        if var.type != expression_type:
            raise_invalid_existential_var(
                'defining_expression_type_mismatch', node)

        var.set_ref(right, self.viper.Old(right, right.pos(), right.info()))

    def translate_get_ghost_output(
            self, node: ast.Assign, ctx: Context) -> List[Stmt]:
        """Translate ``GetGhostOutput``."""
        if len(node.targets) != 1:
            raise_invalid_get_ghost_output('multiple_targets', node)
        if not isinstance(node.targets[0], ast.Name):
            raise_invalid_get_ghost_output('target_not_variable', node)
        target_name = cast(ast.Name, node.targets[0]).id
        target = ctx.actual_function.get_variable(target_name)
        assert target

        operation_call, result_name_node = cast(ast.Call, node.value).args

        if not isinstance(result_name_node, ast.Str):
            raise_invalid_get_ghost_output('result_identifier_not_str', node)
        result_name = cast(ast.Str, result_name_node).s

        if not (isinstance(operation_call, ast.Call) and
                isinstance(operation_call.func, ast.Name)):
            raise_invalid_get_ghost_output('argument_not_io_operation', node)
        operation_call = cast(ast.Call, operation_call)
        operation_name = cast(ast.Name, operation_call.func).id

        if operation_name not in ctx.module.io_operations:
            raise_invalid_get_ghost_output('argument_not_io_operation', node)
        operation = ctx.module.io_operations[operation_name]

        result = None
        for result in operation.get_results():
            if result.name == result_name:
                break
        else:
            raise_invalid_get_ghost_output('invalid_result_identifier', node)
        assert result

        if result.type != target.type:
            raise_invalid_get_ghost_output('type_mismatch', node)

        if len(operation_call.args) != len(operation.get_parameters()):
            raise_invalid_operation_use('result_used_argument', node)
        getter = self.create_result_getter(operation_call, result, ctx)

        position = self.to_position(node, ctx)
        info = self.no_info(ctx)
        set_defined = self.set_var_defined(target, position, info)
        assignment = self.viper.LocalVarAssign(target.ref(), getter,
                                               position, info)

        return [assignment, set_defined]

    def translate_io_contractfunc_call(self, node: ast.Call,
                                       ctx: Context) -> StmtsAndExpr:
        """Translate a call to a IO contract function.

        Currently supported functions:

        +   ``token``
        +   ``ctoken``
        +   ``Open``
        """
        func_name = get_func_name(node)
        if func_name == 'token':
            return self.translate_must_invoke_token(node, ctx)
        elif func_name == 'ctoken':
            return self.translate_must_invoke_ctoken(node, ctx)
        elif func_name == 'Open':
            return self._translate_open(node, ctx)
        elif func_name == 'Eval':
            return self._translate_eval(node, ctx)
        elif func_name == 'eval_io':
            return self._translate_eval_io(node, ctx)
        else:
            raise UnsupportedException(node,
                                       'Unsupported contract function.')

    def _translate_eval_io(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a reference to the eval_io IOOperation as usual. Also stores the
        information that given argument function is among the argument functions used with
        eval_io in the program; this information is used to generate additional return
        type postconditions for the result getter of eval_io.
        """
        if not isinstance(node.args[1], ast.Name):
            raise InvalidProgramException(node, 'invalid.eval.function')
        target = self.get_target(node.args[1], ctx)
        if (ctx.current_function.name != 'Eval' and
                (not isinstance(target, PythonMethod) or not target.pure)):
            raise InvalidProgramException(node, 'invalid.eval.function')
        stmt, res = self.translate_io_operation_call(node, ctx)
        func_stmt, func_arg = self.translate_expr(node.args[1], ctx)
        if func_stmt:
            raise InvalidProgramException(node, 'purity.violated')
        eval_io = self.get_target(node, ctx)

        arg_type = self.get_type(node.args[1], ctx)
        if arg_type.name != CALLABLE_TYPE:
            eval_io.func_args.append((func_arg, arg_type))
        return stmt, res

    def _translate_eval(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the Eval method (which implements the eval_io IOOperation).
        The translation is like for other method calls, except we also inhale that the
        result of the operation is equal to the given function applied to the given
        argument.
        """
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)
        if not isinstance(node.args[1], ast.Name):
            raise InvalidProgramException(node, 'invalid.eval.function')
        target = self.get_target(node.args[1], ctx)
        if not target.pure:
            raise InvalidProgramException(node, 'invalid.eval.function')
        arg_stmt, arg = self.translate_expr(node.args[2], ctx)
        if arg_stmt:
            raise InvalidProgramException(node, 'purity.violated')
        arg_type = self.get_type(node.args[2], ctx)
        func_stmt, func_val = self.translate_normal_call(target, [], [arg], [arg_type],
                                                         node, ctx)
        if func_stmt:
            raise InvalidProgramException(node, 'purity.violated')
        call_stmt, call = self.translate_normal_call_node(node, ctx)
        args = []
        for arg in node.args:
            stmt, arg_val = self.translate_expr(arg, ctx)
            assert not stmt
            args.append(arg_val)

        eval_io = self._get_eval_io_operation(ctx)
        getter = self.create_result_getter(node, eval_io.get_results()[0], ctx, args,
                                           eval_io)
        assume = self.viper.Inhale(self.viper.EqCmp(func_val, getter, position, info),
                                   position, info)
        return call_stmt + [assume], call

    def _get_eval_io_operation(self, ctx: Context) -> PythonIOOperation:
        for module in ctx.module.from_imports:
            if module.type_prefix == 'nagini_contracts.io_builtins':
                io_builtins = module
                break
        else:
            contracts = ctx.module.namespaces['nagini_contracts']
            io_builtins = contracts.namespaces['io_builtins']
        eval_io = io_builtins.io_operations['eval_io']
        return eval_io

    def _translate_open(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """Translate ``Open(io_operation)``."""
        assert ctx.actual_function
        full_perm = self._construct_full_perm(node, ctx)
        opener = Opener(node, ctx, self, full_perm)
        return opener.translate()

    def _translate_results(
            self, operation: PythonIOOperation, node: ast.Call,
            ctx: Context) -> List[Expr]:
        """Translate IO operation results.

        That is: define getters corresponding to operation results or
        emit equalities between each result and getter definition.
        """
        result_translator = ResultTranslator(operation, node, self, ctx)
        return result_translator.translate()

    def _construct_full_perm(self, node: ast.Call,
                             ctx: Context) -> 'viper_ast.FullPerm':
        """Construct silver full perm AST node."""
        return self.viper.FullPerm(self.to_position(node, ctx),
                                   self.no_info(ctx))



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Permission typed Silver expressions."""


from nagini_translation.lib.program_nodes import (
    PythonVar,
)
from nagini_translation.lib.silver_nodes.call import CallMixin
from nagini_translation.lib.silver_nodes.expression import Expression
from nagini_translation.lib.silver_nodes.perm_cmp_expr import (
    PermGtCmp,
    PermGeCmp,
    PermLtCmp,
    PermLeCmp,
)
from nagini_translation.lib.silver_nodes.types import PERM
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Position,
)


class PermExpression(Expression):   # pylint: disable=abstract-method
    """A base class for all perm typed expressions."""

    def __sub__(self, other: 'PermExpression') -> 'PermExpression':
        return PermSub(self, other)

    def __gt__(self, other: 'PermExpression') -> 'BoolExpression':
        return PermGtCmp(self, other)

    def __ge__(self, other: 'PermExpression') -> 'BoolExpression':
        return PermGeCmp(self, other)

    def __lt__(self, other: 'PermExpression') -> 'BoolExpression':
        return PermLtCmp(self, other)

    def __le__(self, other: 'PermExpression') -> 'BoolExpression':
        return PermLeCmp(self, other)


class RawPermExpression(PermExpression):
    """A perm typed Silver expression."""

    def __init__(self, expression: Expr) -> None:
        self._expression = expression

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        assert self._expression.typ() == translator.viper.Perm
        return self._expression


class PermCall(PermExpression, CallMixin):
    """A call to a perm function.

    For example ``foo()`` where ``foo`` is defined as:

    .. code:: silver
        function foo(): Perm
    """

    def __init__(self, function: str, args: 'List[CallArg]') -> None:
        super().__init__(function, args, PERM)


class PermSub(PermExpression):
    """A subtraction of two permission values."""

    def __init__(self, left: PermExpression, right: PermExpression) -> None:
        self._left = left
        self._right = right

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        left = self._left.translate(translator, ctx, position, info)
        right = self._right.translate(translator, ctx, position, info)
        return translator.viper.PermSub(
            left, right, position, info)


class NoPerm(PermExpression):
    """No permission."""

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        return translator.viper.NoPerm(position, info)


class FullPerm(PermExpression):
    """Full permission."""

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        return translator.viper.FullPerm(position, info)


class WildcardPerm(PermExpression):
    """Full permission."""

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        return translator.viper.WildcardPerm(position, info)


class CurrentPerm(PermExpression):
    """The current permission amount to a predicate."""

    def __init__(self, location: 'Location') -> None:
        self._location = location

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        location = self._location.translate(translator, ctx, position, info)
        return translator.viper.CurrentPerm(
            location, position, info)


class IntegerPerm(PermExpression):
    """A multiplication of full permission."""

    def __init__(self, value: 'IntExpression') -> None:
        self._value = value

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        value = self._value.translate(translator, ctx, position, info)
        full_perm = translator.viper.FullPerm(position, info)
        return translator.viper.IntPermMul(
            value, full_perm, position, info)


class PermVar(PermExpression):
    """A reference to a Perm-typed variable."""

    def __init__(self, var: PythonVar) -> None:
        self._var = var

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        assert self._var.decl.typ() == translator.viper.Perm
        return self._var.ref()



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast
import imp
import os
import sys
import types

from astor import to_source
from nagini_contracts.transformer import transform_ast

class ContractsImporter(object):
    """
    Import hook for use on `sys.meta_path`.
    """

    def __init__(self):
        self._cache = {}

    def find_module(self, name, path=None):
        try:
            lastname = name.split('.')[-1]
            # print("find_module", name, path)
            self._cache[name] = imp.find_module(lastname, path), path
        except ImportError:
            return None
        return self

    def load_module(self, name):
        # print("load_module", name)
        try:
            (fd, fn, info), path = self._cache[name]
        except KeyError:
            # can that happen?
            raise ImportError(name)
        if info[2] == imp.PY_SOURCE:
            newpath = None
            filename = fn
            with fd:
                code = fd.read()
        elif info[2] == imp.PY_COMPILED:
            newpath = None
            filename = fn[:-1]
            with open(filename, 'U') as f:
                code = f.read()
        elif info[2] == imp.PKG_DIRECTORY:
            filename = os.path.join(fn, '__init__.py')
            newpath = [fn]
            with open(filename, 'U') as f:
                code = f.read()
        else:
            return imp.load_module(name, fd, fn, info)
        try:
            module = types.ModuleType(name)
            module.__file__ = filename
            if newpath:
                module.__path__ = newpath
            tree = ast.parse(code)
            # print("Before Transformation:\n%s" % to_source(tree))
            transformed_ast = transform_ast(tree)
            # print("After Transformation:\n%s" % to_source(transformed_ast))
            code = compile(transformed_ast, filename, 'exec')
            sys.modules[name] = module
            exec(code, module.__dict__)
            return module
        except Exception:
            raise ImportError('cannot import %s' % (name))


def install_hook():
    print("Installing import hook.")
    sys.meta_path.insert(0, ContractsImporter())


def remove_hook():
    print("Removing import hook.")
    sys.meta_path[:] = [importer for importer in sys.meta_path if
                        not isinstance(importer, ContractsImporter)]



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Analyzer that collects information about IO operations."""


import ast

from mypy.types import AnyType
from nagini_contracts.io_contracts import IO_OPERATION_PROPERTY_FUNCS
from typing import cast, List

from nagini_translation.lib import program_nodes as nodes
from nagini_translation.lib.constants import (
    BOOL_TYPE,
    EVAL_IO_SIGNATURE,
    OBJECT_TYPE,
)
from nagini_translation.lib.util import (
    construct_lambda_prefix,
    InvalidProgramException,
    UnsupportedException,
)


class IOOperationAnalyzer(ast.NodeVisitor):
    """Walks through IO operation AST and collects the needed information."""

    def __init__(
            self, parent: 'nagini_translation.analyzer.Analyzer',
            node_factory: nodes.ProgramNodeFactory) -> None:
        self._parent = parent
        self._types = parent.types
        self._node_factory = node_factory
        self._place_class = parent.find_or_create_class(
            'Place', module=parent.module.global_module)
        self._place_class.superclass = parent.find_or_create_class(
            'object', module=parent.module.global_module)

        self._current_io_operation = None   # type: nodes.PythonIOOperation
        self._current_node = None           # type: ast.FunctionDef
        self._in_property = False

    def _raise_invalid_operation(
            self,
            error_type: str,
            node: ast.AST = None):
        """Raise InvalidProgramException."""
        node = node or self._current_node
        raise InvalidProgramException(
            node,
            'invalid.io_operation.' + error_type,
        )

    def _create_io_operation(
            self, node: ast.FunctionDef) -> nodes.PythonIOOperation:
        """Create IO operation.

        Creates non-initialized IO operation from an AST node and adds
        it to the module.
        """
        name = node.name
        assert isinstance(name, str)
        operation = self._node_factory.create_python_io_operation(
            name,
            node,
            self._parent.module,
            self._node_factory,
        )
        self._parent.module.io_operations[name] = operation
        return operation

    def _check_type(self) -> None:
        """Check if operation type is ``bool``."""
        op_name = self._current_io_operation.name
        func_type = self._parent.module.get_func_type([op_name])
        if isinstance(func_type, AnyType):
            self._raise_invalid_operation('return_type_not_bool')
        operation_type = self._parent.convert_type(func_type,
                                                   self._current_io_operation.node)
        if not operation_type or operation_type.name != BOOL_TYPE:
            self._raise_invalid_operation('return_type_not_bool')

    def _check_arg_types(self) -> None:
        """Allow only certain types of args.

        Do not allow ``*args`` and ``**kwargs``. Only ``Result()`` is
        allowed as default value.
        """
        node = self._current_node
        if node.args.vararg:
            self._raise_invalid_operation('vararg')
        if node.args.kwarg:
            self._raise_invalid_operation('kwarg')
        for default in node.args.defaults:
            if (not isinstance(default, ast.Call) or
                    not isinstance(default.func, ast.Name) or   # type: ignore
                    not default.func.id == 'Result'):  # type: ignore
                self._raise_invalid_operation('default_argument')

    def _typeof(self, node: ast.AST,
                lambda_: ast.Lambda = None) -> nodes.PythonType:
        """Get the type of the given AST node."""
        assert isinstance(node, ast.arg)
        if (self._current_io_operation.name == EVAL_IO_SIGNATURE[0] and
                node.arg == EVAL_IO_SIGNATURE[2]):
            return self._parent.module.global_module.classes[OBJECT_TYPE]
        scopes = [self._current_io_operation.name]
        if lambda_:
            prefix = construct_lambda_prefix(
                lambda_.lineno, lambda_.col_offset)
            scopes.append(prefix)
        typ, _ = self._parent.module.get_type(scopes, node.arg)
        # to avoid problems with boxed versions not being reference-equal,
        # always use unboxed versions
        result = self._parent.convert_type(typ, node).try_unbox()
        return result

    def _set_preset(self, inputs: List[ast.arg]) -> List[ast.arg]:
        """Check and set preset.

        Checks that exactly one place is in preset, sets operation
        preset and returns input list with all places removed.
        """
        if self._current_io_operation.name == 'join_io':
            # Special handling of built-in “join_io”.
            assert len(inputs) == 2
            in_places = []
            for i in range(2):
                assert self._typeof(inputs[i]) == self._place_class
                in_place = self._node_factory.create_python_var(
                    inputs[i].arg, inputs[i], self._place_class)
                in_places.append(in_place)
            self._current_io_operation.set_preset(in_places)
            return []
        if not inputs or self._typeof(inputs[0]) != self._place_class:
            self._raise_invalid_operation('invalid_preset')
        for inp in inputs[1:]:
            if self._typeof(inp) == self._place_class:
                self._raise_invalid_operation('invalid_preset')
        in_place = self._node_factory.create_python_var(
            inputs[0].arg, inputs[0], self._place_class)
        self._current_io_operation.set_preset([in_place])
        return inputs[1:]

    def _set_postset(self, outputs: List[ast.arg]) -> List[ast.arg]:
        """Check and set postset.

        Checks that exactly one place is in postset, sets operation
        postset and returns input list with all places removed.
        """
        counter = 0
        for output in outputs:
            if self._typeof(output) == self._place_class:
                counter += 1
        if counter > 1:
            if self._current_io_operation.name == 'split_io':
                # Special handling of built-in “split_io”.
                assert len(outputs) == 2
                out_places = []
                for i in range(2):
                    assert self._typeof(outputs[i]) == self._place_class
                    out_place = self._node_factory.create_python_var(
                        outputs[i].arg, outputs[i], self._place_class)
                    out_places.append(out_place)
                self._current_io_operation.set_postset(out_places)
                return []
            else:
                self._raise_invalid_operation('invalid_postset')
        elif counter == 1:
            if self._typeof(outputs[-1]) != self._place_class:
                self._raise_invalid_operation('invalid_postset')
            out_place = self._node_factory.create_python_var(
                outputs[-1].arg, outputs[-1], self._place_class)
            self._current_io_operation.set_postset([out_place])
            return outputs[:-1]
        else:
            return outputs

    def _parse_arguments(self) -> None:
        """Parse and check operation arguments."""
        node = self._current_node

        self._check_arg_types()

        if node.args.defaults:
            inputs = node.args.args[:-len(node.args.defaults)]
            outputs = node.args.args[-len(node.args.defaults):]
        else:
            inputs = node.args.args[:]
            outputs = []

        inputs = self._set_preset(inputs)
        outputs = self._set_postset(outputs)

        self._current_io_operation.set_inputs([
            self._node_factory.create_python_var(
                node.arg, node, self._typeof(node))
            for node in inputs])
        self._current_io_operation.set_outputs([
            self._node_factory.create_python_var(
                node.arg, node, self._typeof(node))
            for node in outputs])

    def analyze_io_operation(self, node: ast.FunctionDef) -> None:
        """Analyze AST node representing IO operation.

        That is: extract information and perform some basic
        well-formedness checks.
        """
        assert self._current_io_operation is None
        assert self._current_node is None

        self._current_node = node
        operation = self._create_io_operation(node)
        self._current_io_operation = operation
        self._check_type()
        self._parse_arguments()

        for child in node.body:
            self.visit(child)

        self._current_node = None
        self._current_io_operation = None

    def visit_Return(self, node: ast.Return) -> None:
        """Parse IO operation body.

        IO operation body must be a single expression that is returned.
        """
        body = node.value

        if (isinstance(body, ast.Call) and
                isinstance(body.func, ast.Call) and
                isinstance(body.func.func, ast.Name) and
                body.func.func.id.startswith('IOExists')):
            lambda_ = cast(ast.Lambda, body.args[0])
            args = lambda_.args.args
            io_existential_creators = [
                self._node_factory.create_python_var_creator(
                    arg.arg, arg, self._typeof(arg, lambda_))
                for arg in args
            ]
            body = lambda_.body
        else:
            io_existential_creators = []
        if not self._current_io_operation.set_body(body):
            self._raise_invalid_operation(
                'duplicate_body',
                node)
        assert self._current_io_operation.set_io_existentials(
            io_existential_creators)

    def visit_Call(self, node: ast.Call) -> None:
        """Parse IO operation properties.

        Currently, only parses properties such as ``Terminates`` and
        ``TerminationMeasure``.
        """
        assert self._current_io_operation is not None
        assert self._current_node is not None

        if (isinstance(node.func, ast.Name) and
                node.func.id in IO_OPERATION_PROPERTY_FUNCS):

            for child in self._current_node.body:
                if (isinstance(child, ast.Expr) and
                        child.value == node):
                    break
            else:
                self._raise_invalid_operation(
                    'misplaced_property',
                    node)

            operation = self._current_io_operation
            arg = node.args[0]
            if node.func.id == 'Terminates':
                if not operation.set_terminates(arg):
                    self._raise_invalid_operation(
                        'duplicate_property',
                        node)
            elif node.func.id == 'TerminationMeasure':
                if not operation.set_termination_measure(arg):
                    self._raise_invalid_operation(
                        'duplicate_property',
                        node)
            else:
                raise UnsupportedException(node,
                                           'Unsupported property type.')
            self._in_property = True
            for arg in node.args:
                self.visit(arg)
            self._in_property = False
        else:
            for arg in node.args:
                self.visit(arg)

    def visit_Name(self, node: ast.Name) -> None:
        """Check if node is an operation input.

        This method is expected to be called by ``visit_Call`` on IO
        operation property arguments. Checks if provided node is an
        operation input and raises an error, if it is not.
        """
        if self._in_property:
            if not self._current_io_operation.is_input(node.id):
                self._raise_invalid_operation('depends_on_not_imput', node)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from nagini_translation.lib.program_nodes import PythonMethod
from nagini_translation.lib.typedefs import (
    Expr,
    Stmt,
    StmtsAndExpr,
)
from nagini_translation.lib.util import (
    InvalidProgramException,
    UnsupportedException,
    get_func_name
)
from nagini_translation.translators.abstract import Context
from nagini_translation.translators.common import CommonTranslator


class PermTranslator(CommonTranslator):

    def translate_perm(self, node: ast.AST, ctx: Context) -> Expr:
        """
        Generic visitor function for translating a permission amount
        """
        method = 'translate_perm_' + node.__class__.__name__
        visitor = getattr(self, method, self.translate_generic)
        return visitor(node, ctx)

    def translate_perm_Num(self, node: ast.Num, ctx: Context) -> Expr:
        if node.n == 1:
            return self.viper.FullPerm(self.to_position(node, ctx),
                                       self.no_info(ctx))
        raise UnsupportedException(node)

    def translate_perm_or_int(self, node: ast.AST, ctx: Context):
        if isinstance(node, ast.Num):
            stmt, int_val = self.translate_expr(node, ctx, self.viper.Int)
            if stmt:
                raise InvalidProgramException(node, 'purity.violated')
            return int_val, True
        else:
            int_or_perm_val = self.translate_perm(node, ctx)
            return int_or_perm_val, int_or_perm_val.typ() == self.viper.Int

    def translate_perm_BinOp(self, node: ast.BinOp, ctx: Context) -> Expr:

        if isinstance(node.op, ast.Div):
            left, left_int = self.translate_perm_or_int(node.left, ctx)
            right_stmt, right = self.translate_expr(node.right, ctx,
                                                    target_type=self.viper.Int)
            if right_stmt:
                raise InvalidProgramException(node, 'purity.violated')

            if left_int:
                return self.viper.FractionalPerm(left, right,
                                                 self.to_position(node, ctx),
                                                 self.no_info(ctx))
            return self.viper.PermDiv(left, right,
                                      self.to_position(node, ctx),
                                      self.no_info(ctx))

        if isinstance(node.op, ast.Mult):
            left, left_int = self.translate_perm_or_int(node.left, ctx)
            right, right_int = self.translate_perm_or_int(node.right, ctx)

            if left_int != right_int and right_int:
                left, left_int, right, right_int = right, right_int, left, left_int

            if left_int and right_int:
                return self.viper.Mul(left, right,
                                      self.to_position(node, ctx),
                                      self.no_info(ctx))
            if left_int or right_int:
                return self.viper.IntPermMul(left, right,
                                             self.to_position(node, ctx),
                                             self.no_info(ctx))

            return self.viper.PermMul(left, right,
                                      self.to_position(node, ctx),
                                      self.no_info(ctx))

        new_node = None
        if isinstance(node.op, ast.Add):
            new_node = self.viper.PermAdd
        elif isinstance(node.op, ast.Sub):
            new_node = self.viper.PermSub

        if new_node:
            left = self.translate_perm(node.left, ctx)
            right = self.translate_perm(node.right, ctx)
            return new_node(left, right,
                            self.to_position(node, ctx),
                            self.no_info(ctx))
        raise UnsupportedException(node)

    def translate_perm_Call(self, node: ast.Call, ctx: Context) -> Expr:
        func_name = get_func_name(node)
        if func_name == 'ARP':
            if not ctx.arp:
                raise UnsupportedException(node, 'ARP not supported. Use --arp flag.')
            if len(node.args) == 0:
                return self.get_arp_for_context(node, ctx)
            elif len(node.args) == 1:
                arg0_stmt, arg0 = self.translate_expr(node.args[0], ctx, self.viper.Int)
                if arg0_stmt:
                    raise InvalidProgramException(node, 'purity.violated')
                formal_arg = self.viper.LocalVarDecl('count', self.viper.Int,
                                                     self.to_position(node, ctx),
                                                     self.no_info(ctx))
                return self.viper.FuncApp('rdc', [arg0], self.to_position(node, ctx),
                                          self.no_info(ctx), self.viper.Perm,
                                          [formal_arg])
        elif func_name == 'getARP':
            if not ctx.arp:
                raise UnsupportedException(node, 'ARP not supported. Use --arp flag.')
            if len(node.args) == 1:
                formal_arg = self.viper.LocalVarDecl(
                    'tk', self.viper.Ref, self.to_position(node, ctx), self.no_info(ctx))
                arg0_stmt, arg0 = self.translate_expr(node.args[0], ctx, self.viper.Ref)
                if arg0_stmt:
                    raise InvalidProgramException(node, 'purity.violated')
                return self.viper.FuncApp('rd_token', [arg0], self.to_position(node, ctx),
                                          self.no_info(ctx), self.viper.Perm,
                                          [formal_arg])

        call_stmt, call = self.translate_expr(node, ctx, self.viper.Int)
        if not call_stmt:
            return call
        raise InvalidProgramException(node, 'purity.violated')

    def translate_perm_Name(self, node: ast.Name, ctx: Context) -> Expr:
        if node.id == 'RD_PRED':
            if not ctx.arp:
                raise UnsupportedException(node, 'ARP not supported. Use --arp flag.')
            return self.viper.FuncApp('globalRd', [], self.to_position(node, ctx),
                                      self.no_info(ctx), self.viper.Perm, {})
        stmt, res = self.translate_expr(node, ctx)
        if stmt:
            raise InvalidProgramException(node, 'purity.violated')
        return  res

    def translate_perm_Attribute(self, node: ast.Attribute, ctx: Context) -> Expr:
        stmt, expr = self.translate_expr(node, ctx, self.viper.Int)
        if not stmt:
            return expr
        raise InvalidProgramException(node, 'purity.violated')

    def get_arp_for_context(self, node: ast.AST, ctx: Context):
        if (ctx.actual_function and isinstance(ctx.actual_function, PythonMethod) and
                ctx.actual_function.pure):
            return self.viper.WildcardPerm(self.to_position(node, ctx), self.no_info(ctx))
        else:
            if not ctx.arp:
                raise UnsupportedException(node, 'ARP not supported. Use --arp flag.')
            if ctx.current_thread_object is not None:
                formal_arg = self.viper.LocalVarDecl('tk', self.viper.Ref,
                                                     self.to_position(node, ctx),
                                                     self.no_info(ctx))
                if ctx.is_thread_start:
                    return self.viper.FuncApp('rd_token_fresh',
                                              [ctx.current_thread_object],
                                              self.to_position(node, ctx),
                                              self.no_info(ctx), self.viper.Perm,
                                              [formal_arg])
                else:
                    return self.viper.FuncApp('rd_token', [ctx.current_thread_object],
                                              self.to_position(node, ctx),
                                              self.no_info(ctx), self.viper.Perm,
                                              [formal_arg])
            else:
                return self.viper.FuncApp('rd', [], self.to_position(node, ctx),
                                          self.no_info(ctx), self.viper.Perm, {})



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Comparisons of permission typed Silver expressions."""


from nagini_translation.lib.silver_nodes.bool_expr import BoolExpression
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Position,
)


class PermComparison(BoolExpression):   # pylint: disable=abstract-method
    """A base class for all permission comparison operators."""


class PermGtCmp(PermComparison):
    """Greater than permission comparison."""

    def __init__(self, left: 'PermExpression', right: 'PermExpression') -> None:
        self._left = left
        self._right = right

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        left = self._left.translate(translator, ctx, position, info)
        right = self._right.translate(translator, ctx, position, info)
        return translator.viper.PermGtCmp(
            left, right, position, info)


class PermGeCmp(PermComparison):
    """Greater equal permission comparison."""

    def __init__(self, left: 'PermExpression', right: 'PermExpression') -> None:
        self._left = left
        self._right = right

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        left = self._left.translate(translator, ctx, position, info)
        right = self._right.translate(translator, ctx, position, info)
        return translator.viper.PermGeCmp(
            left, right, position, info)


class PermLtCmp(PermComparison):
    """Less than permission comparison."""

    def __init__(self, left: 'PermExpression', right: 'PermExpression') -> None:
        self._left = left
        self._right = right

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        left = self._left.translate(translator, ctx, position, info)
        right = self._right.translate(translator, ctx, position, info)
        return translator.viper.PermLtCmp(
            left, right, position, info)


class PermLeCmp(PermComparison):
    """Less equal permission comparison."""

    def __init__(self, left: 'PermExpression', right: 'PermExpression') -> None:
        self._left = left
        self._right = right

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        left = self._left.translate(translator, ctx, position, info)
        right = self._right.translate(translator, ctx, position, info)
        return translator.viper.PermLeCmp(
            left, right, position, info)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Base classes for obligations."""


import abc
import ast

from typing import Any, Dict, List, Optional, Tuple

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.context import Context
from nagini_translation.lib.errors import Rules
from nagini_translation.lib.program_nodes import (
    PythonMethod,
)
from nagini_translation.lib.typedefs import (
    Predicate,
    Field,
)
from nagini_translation.translators.common import CommonTranslator


class ObligationInstance(abc.ABC):
    """A base class for all obligation instances.

    Obligation instance is a concrete instantiation of a obligation with
    specific measure and other properties.
    """

    def __init__(self, obligation: 'Obligation', node: ast.expr) -> None:
        super().__init__()
        self.obligation = obligation
        """Actual obligation type."""
        self.node = node
        """AST node from which this obligation was instantiated."""

    @abc.abstractmethod
    def is_fresh(self) -> bool:
        """Return if this a fresh obligation."""

    @abc.abstractmethod
    def get_measure(self) -> "IntegerExpression":
        """Return an obligation measure.

        If obligation is fresh, the behaviour of this call is undefined.
        """

    @abc.abstractmethod
    def get_target(self) -> sil.RefExpression:
        """Return an expression to which obligation is attached."""

    @abc.abstractmethod
    def get_use_method(
            self, ctx: Context) -> List[Tuple[sil.Expression, Rules]]:
        """Get inhale exhale pair for use in method contract."""

    @abc.abstractmethod
    def get_use_loop(
            self, ctx: Context) -> List[Tuple[sil.Expression, Rules]]:
        """Get inhale exhale pair for use in loop invariant."""

    @abc.abstractmethod
    def get_obligation_bound(self, ctx: Context) -> sil.Statement:
        """Get exhale/inhale pair that bounds obligation."""


class Obligation(abc.ABC):
    """A base class for all obligations."""

    def __init__(self, predicate_names, field_names) -> None:
        self._predicate_names = predicate_names
        self._field_names = field_names

    @abc.abstractmethod
    def identifier(self) -> str:
        """Unique identifier of this obligation type."""

    @abc.abstractmethod
    def check_node(
            self, node: ast.Call,
            obligation_info: 'PythonMethodObligationInfo',
            method: PythonMethod) -> Optional[ObligationInstance]:
        """Check if node represents this obligation type.

        If check is successful, an obligation instance object is
        returned. Otherwise – ``None``.
        """

    @abc.abstractmethod
    def generate_axiomatized_preconditions(
            self, obligation_info: 'PythonMethodObligationInfo',
            interface_dict: Dict[str, Any]) -> List[sil.BoolExpression]:
        """Generate obligations for axiomatic method precondition."""

    @abc.abstractmethod
    def create_leak_check(self, var_name: str) -> List[sil.BoolExpression]:
        """Create a leak check for this obligation.

        :param var_name: variable name to be used in ``ForPerm``
        """

    def _create_predicate_for_perm(
            self, predicate_name: str, var_name: str) -> sil.ForPerm:
        """Create a ForPerm expression with predicate for use in leak check."""
        return sil.ForPerm(
            var_name,
            sil.Predicate(predicate_name, var_name),
            sil.FalseLit())

    def _create_field_for_perm(
            self, field_name: str, var_name: str) -> sil.ForPerm:
        """Create a ForPerm expression with field for use in leak check."""
        return sil.ForPerm(
            var_name,
            sil.Field(field_name, sil.INT),
            sil.FalseLit())

    def create_predicates(
            self, translator: CommonTranslator) -> List[Predicate]:
        """Create predicates that are used to represent this obligation."""
        position = translator.viper.NoPosition
        info = translator.viper.NoInfo
        predicates = [
            sil.Predicate(name, 'r')
            for name in self._predicate_names]
        translated_predicates = [
            predicate.translate(translator, None, position, info)
            for predicate in predicates]
        return translated_predicates

    def create_fields_untranslated(self) -> List[sil.Field]:
        """Create fields that are used to represent this obligation."""
        fields = [
            sil.Field(name, sil.INT)
            for name in self._field_names]
        return fields

    def create_fields(
            self, translator: CommonTranslator) -> List[Field]:
        """Create Silver fields that are used to represent this obligation."""
        position = translator.viper.NoPosition
        info = translator.viper.NoInfo
        translated_fields = [
            field.translate(translator, None, position, info)
            for field in self.create_fields_untranslated()]
        return translated_fields
